import os
from time import time
import pymatgen.analysis.structure_matcher as sm
from pyxtal import pyxtal
from pyxtal.interface.charmm import CHARMM
from pyxtal.interface.dftb import DFTB, DFTB_relax
from pyxtal.interface.gulp import GULP_OC as GULP_relax
from pyxtal.msg import ReadSeedError
from pyxtal.representation import representation
from pyxtal.util import ase2pymatgen


class benchmark:
    """
    Automated benchmark for experimental crystals with
        - VASP
        - DFTB
        - ANI
        - GULP
        - CHARMM
    """

    def __init__(self, struc, smiles, clean=True, work_dir="tmp", **kwargs):
        """
        Args:
            struc: pymatgen structure or filename
            smiles: list of smile codes
            skf_dir (optional): DFTB parameter directory
            charmm_info (optional): list of atom labels
            charmm_prm (optional): charmm prm file
            charmm_rtf (optional): charmm rtf file
            gulp_info (optional): gulp charge/label info
            clean: whether or not clean up tmp files
        """
        self.clean = clean
        self.valid = True
        self.work_dir = work_dir
        if not os.path.exists(work_dir):
            os.makedirs(work_dir)

        for i in range(len(smiles)):
            if not smiles[i].endswith(".smi"):
                smiles[i] = smiles[i] + ".smi"
        self.smiles = smiles

        xtal = pyxtal(molecular=True)
        try:
            try:
                xtal.from_seed(struc, molecules=smiles)
            except ReadSeedError:
                xtal.from_seed(struc, molecules=smiles, add_H=True)

            # Check if it needs to transform to subgroup representation
            if xtal.has_special_site():
                xtal = xtal.to_subgroup()
            # initialize dicts
            self.xtal = {}
            self.rep = {}
            self.time = {}
            self.energy = {}
            self.pmg = {}
            self.ase = {}
            self.diff = {}
            self.Z = sum(xtal.numMols)
            self.xtal["reference"] = xtal
            self.rep["reference"] = representation.from_pyxtal(xtal).x
            self.ase["reference"] = xtal.to_ase(resort=True)
            self.pmg["reference"] = xtal.to_pymatgen()
            self.pmg["reference"].remove_species("H")  # remove hydrogen
        except ReadSeedError:
            print("Fail to read crystal")
            self.valid = False

        # set up the optional parameters
        if "skf_dir" in kwargs:
            self.skf_dir = kwargs.pop("skf_dir")

        if "charmm_info" in kwargs:
            self.charmm_info = kwargs.pop("charmm_info")

        if "charmm_prm" in kwargs:
            self.charmm_prm = kwargs.pop("charmm_prm")

        if "charmm_rtf" in kwargs:
            self.charmm_rtf = kwargs.pop("charmm_rtf")

        if "gulp_info" in kwargs:
            self.gulp_info = kwargs.pop("gulp_info")

    def calc(self, calculator, show=True):
        """
        Execute calculation
        """

        cwd = os.getcwd()
        os.chdir(self.work_dir)
        if calculator.find("dftb") > -1:
            tmp = calculator.split("_")
            disp = tmp[-1] if len(tmp) > 1 else "D3"
            self.dftb(disp, show)
        elif calculator == "vasp":
            self.vasp(show)
        elif calculator == "ani":
            self.ani(show)
        elif calculator == "charmm":
            self.charmm(show)
        elif calculator == "gulp":
            self.gulp(show)
        else:
            raise KeyError("unknown calculator", calculator)

        os.chdir(cwd)

    def dftb(self, disp, show=True, logfile="ase.log"):
        """
        DFTB calculation
        """
        if not hasattr(self, "skf_dir"):
            raise KeyError("no skf_dir for DFTB calculation")
        t0 = time()
        ase = self.ase["reference"].copy()
        # ase = dftb_relax(ase, self.skf_dir, kresol=0.08, logfile=logfile)
        ase, _ = DFTB(ase, self.skf_dir, mode="relax",
                      kresol=0.08, disp=disp)
        ase, _ = DFTB(ase, self.skf_dir, mode="vc-relax",
                      step=300, kresol=0.06, disp=disp)
        ase = DFTB_relax(ase, self.skf_dir, opt_cell=True,
                         kresol=0.06, disp=disp, logfile=logfile)
        xtal = pyxtal(molecular=True)
        pmg = ase2pymatgen(ase)
        xtal.from_seed(pmg, molecules=self.smiles)

        pmg.remove_species("H")
        calc = "dftb_" + disp

        self.xtal[calc] = xtal
        self.pmg[calc] = pmg
        self.energy[calc] = ase.get_potential_energy()
        self.rep[calc] = representation.from_pyxtal(xtal).x
        self.time[calc] = time() - t0
        if show: self.summary(calc)

    def vasp(self, show=True):
        """
        VASP calculation
        """
        t0 = time()
        ase, energy = vasp_relax(self.ase["reference"].copy())
        ase, energy = vasp_relax(ase, opt_cell=True)
        xtal = pyxtal(molecular=True)
        pmg = ase2pymatgen(ase)
        xtal.from_seed(pmg, molecules=self.smiles)

        pmg.remove_species("H")
        self.xtal["vasp"] = xtal
        self.pmg["vasp"] = pmg
        self.rep["vasp"] = representation.from_pyxtal(xtal).x
        self.energy["vasp"] = energy
        self.time["vasp"] = time() - t0
        if show:
            self.summary("vasp")

    def ani(self, show=True, logfile="ase-ani.log"):
        """
        Torch ANI calculation
        """

        t0 = time()
        # self.ase.write('ani.cif', format='cif')
        ase = self.ase["reference"].copy()
        ase = ani_relax(ase, logfile=logfile)
        ase = ani_relax(ase, opt_cell=True, logfile=logfile, max_time=10.0)
        ase = ani_relax(ase, opt_cell=True, logfile=logfile, max_time=10.0)

        # ase.write('ani_final.cif', format='cif'); import sys; sys.exit()
        xtal = pyxtal(molecular=True)
        self.ase["ani"] = ase
        pmg = ase2pymatgen(ase)
        try:
            xtal.from_seed(pmg, molecules=self.smiles)
            pmg.remove_species("H")
            self.xtal["ani"] = xtal
            self.pmg["ani"] = pmg
            self.rep["ani"] = representation.from_pyxtal(xtal).x
            self.energy["ani"] = ase.get_potential_energy()
            self.time["ani"] = time() - t0
            if show: self.summary("ani")
        except ReadSeedError:
            print("Molecular form is broken after relaxation")

    def charmm(self, show=True, steps=None):
        """
        CHARMM-GAFF
        """
        if steps is None:
            steps = [2000, 3000]
        struc = self.xtal["reference"].copy()

        t0 = time()
        calc = CHARMM(struc, "ben", steps=steps,
                      atom_info=self.charmm_info,
                      debug=True)
        calc.run(clean=self.clean)  #False); import sys; sys.exit()
        struc = calc.structure

        pmg = struc.to_pymatgen()
        pmg.remove_species("H")
        self.xtal["charmm"] = struc
        self.pmg["charmm"] = pmg
        self.rep["charmm"] = representation.from_pyxtal(struc).x
        self.energy["charmm"] = calc.structure.energy
        self.time["charmm"] = time() - t0
        if show:
            self.summary("charmm")

    def gulp(self, show=True, step=None, stepmx=None):
        """
        GULP-GAFF
        """
        if stepmx is None:
            stepmx = [0.001, 0.005, 0.02]
        if step is None:
            step = [400, 400, 1000]
        struc = self.xtal["reference"].copy()
        g_info = self.gulp_info
        t0 = time()
        calc = GULP_relax(struc, "ben", opt="conv", steps=step[0],
                          stepmx=stepmx[0], atom_info=g_info)
        calc.run(clean=self.clean)  # ; print(os.getcwd()); import sys; sys.exit()
        if not calc.optimized:
            raise RuntimeError("GULP calculation is wrong")
        struc = calc.structure
        calc = GULP_relax(
            struc,
            "ben",
            opt="conp",
            steps=step[1],
            stepmx=stepmx[1],
            atom_info=g_info,
            dump="1.cif",
        )
        calc.run(clean=self.clean)
        if not calc.optimized:
            raise RuntimeError("GULP calculation is wrong")
        struc = calc.structure
        calc = GULP_relax(struc, "ben", opt="conp", steps=step[2],
                          stepmx=stepmx[2], atom_info=g_info)
        # print(struc)
        calc.run(clean=self.clean) #, pause=True); import sys; sys.exit()
        struc = calc.structure
        if not calc.optimized:
            raise RuntimeError("GULP calculation is wrong")
        pmg = struc.to_pymatgen()
        pmg.remove_species("H")
        self.xtal["gulp"] = struc
        self.pmg["gulp"] = pmg
        self.rep["gulp"] = representation.from_pyxtal(struc).x
        self.energy["gulp"] = calc.structure.energy
        self.time["gulp"] = time() - t0
        if show:
            self.summary("gulp")

    def charmm_md(self):
        raise NotImplementedError

    def summary(self, calc=None):
        calcs = self.time.keys() if calc is None else [calc]

        pmg0 = self.pmg["reference"]
        v0 = pmg0.volume
        matcher = sm.StructureMatcher(ltol=0.3, stol=0.3, angle_tol=10)
        for calc in calcs:
            time = self.time[calc] / 60
            rep = representation(self.rep[calc], self.smiles)
            dv = (self.pmg[calc].volume - v0) / v0
            strs = f"{rep.to_string(eng=self.energy[calc] / self.Z):s} "
            strs += f"{calc:8s} {time:6.2f} {dv:6.3f}"
            rmsd = matcher.get_rms_dist(pmg0, self.pmg[calc])
            if rmsd is not None:
                strs += f"{rmsd[0]:6.3f}{rmsd[1]:6.3f}"
                self.diff[calc] = [dv, rmsd[0], rmsd[1]]
            else:
                self.diff[calc] = [dv, None, None]
            print(strs)


if __name__ == "__main__":
    import warnings
    from pyxtal.db import database
    warnings.filterwarnings("ignore")

    work_dir = "tmp"
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    # if 'DFTB_PREFIX' in os.environ.keys():
    #    skf_dir = os.environ['DFTB_PREFIX'] + '3ob-3-1/'
    # else:
    #    raise RuntimeError("Cannot find DFTB_PREFIX in the environment")
    skf_dir = None

    db = database("pyxtal/database/test.db")
    code = "ACSALA"  #'BENZEN'
    row = db.get_row(code)
    xtal = db.get_pyxtal(code)

    c_info = row.data["charmm_info"]
    with open(work_dir + "/pyxtal.prm", "w") as prm:
        prm.write(c_info["prm"])
    with open(work_dir + "/pyxtal.rtf", "w") as rtf:
        rtf.write(c_info["rtf"])
    g_info = row.data["gulp_info"]

    pmg = xtal.to_pymatgen()
    smi = row.mol_smi.split(".")
    ben = benchmark(
        pmg,
        smi,
        charmm_info=c_info,
        gulp_info=g_info,
        work_dir=work_dir,
        skf_dir=skf_dir,
        clean=False,
    )

    rep = representation(ben.rep["reference"], smi)
    print(rep.to_string() + " reference")

    # for calc in ['charmm', 'ani', 'gulp']: #, 'dftb_TS']:
    # for calc in ['dftb_TS']:
    # for calc in ['ani', 'ani', 'ani', 'ani']:
    for calc in ["charmm", "gulp"]:  # , 'dftb_TS']:
        ben.calc(calc, show=True)
    print("=========================================")
"""
Note: The results are outdated and the codes are obsolete.
qzhu@cms-2:~/GitHub/HT-OCSP$ python htocsp/benchmark.py
29 0 9.20 7.29 6.69 1 0.00 0.75 0.00  111.8 -43.7   19.4 1 reference
29 0 9.20 7.44 6.68 1 0.00 0.75 1.00  -72.0 -42.7   20.6 0     -1.444 charmm  0.15  0.019 0.030 0.032
61 0 6.63 7.42 9.36 1 1.00 0.00 1.00   -7.5  42.5 -108.3 0  -6318.357 ani     0.24  0.027 0.008 0.010
29 0 9.24 7.35 6.64 1 0.00 0.75 1.00  -70.6 -43.2   19.4 0     -0.305 gulp    0.60  0.006 0.024 0.028
61 0 6.56 7.12 9.11 1 0.00 1.00 0.00   -7.8  43.2 -109.0 0   -341.479 dftb_D3 2.31 -0.051 0.017 0.022
61 0 6.70 7.24 9.20 1 0.00 0.00 0.00 -142.9  43.3  -70.8 0   -341.147 dftb_TS 0.82 -0.005 0.011 0.013
"""



import os
import re

import numpy as np
from ase.calculators.calculator import (
    FileIOCalculator,
    kpts2ndarray,
    kpts2sizeandoffsets,
)
from ase.constraints import UnitCellFilter, FixSymmetry
from ase.io import read
from ase.optimize.fire import FIRE
from ase.units import Bohr, Hartree

from pyxtal.util import Kgrid


def make_Hamiltonian(
    skf_dir,
    atom_types,
    disp,
    kpts,
    scc_error=1e-06,
    scc_iter=500,
    write_band=False,
    use_omp=False,
):
    """
    Generate the DFTB Hamiltonian for DFTB+
    """
    if disp == "D3":  # from dftb manual
        dispersion = """DftD3{
                      Damping = BeckeJohnson{
                      a1 = 0.5719
                      a2 = 3.6017
                      }
                      s6 = 1.0
                      s8 = 0.5883
                      }
                      """

    elif disp == "D30":  # zero dampling
        dispersion = """DftD3{
                      Damping = ZeroDamping{
                      sr6 = 0.746
                      alpah6 = 4.191
                      }
                      s6 = 1.0
                      s8 = 3.209
                      }
                     """

    elif disp == "D4":
        dispersion = """DftD4{
                      s6 = 1
                      s8 = 0.6635015
                      s9 = 1
                      a1 = 0.5523240
                      a2 = 4.3537076
                        }
                     """

    elif disp == "MBD":  # 1.0 from J. Phys. Chem. Lett. 2018, 9, 399−405
        dispersion = "MBD{\n\tKGrid = " + str(kpts)[1:-1] + "\n\tBeta = 1.0}\n"

    elif disp == "TS":  # 1.05 from J. Phys. Chem. Lett. 2018, 9, 399−405
        dispersion = """TS{
                      Damping = 20.0
                      RangeSeparation = 1.0
                      }
                     """

    elif disp == "LJ":
        dispersion = "LennardJones{Parameters = UFFParameters{}}"
    else:
        dispersion = None

    kwargs = {
        "Hamiltonian_SCC": "yes",
        "Hamiltonian_SCCTolerance": scc_error,  # 1e-06,
        "Hamiltonian_MaxSCCIterations": scc_iter,  # 1000,
        #'Hamiltonian_Mixer': 'DIIS{}', #Default is Broyden
        #'Hamiltonian_Dispersion': dispersion,
        "slako_dir": skf_dir,
        "Analysis_": "",
        "Analysis_WriteBandOut": "No",
        "Analysis_MullikenAnalysis": "No",
        "Analysis_CalculateForces": "Yes",
        "Analysis_PrintForces": "Yes",
    }
    if write_band:
        kwargs["Analysis_WriteBandOut"] = "Yes"
    if use_omp:
        kwargs["Parallel_"] = ""
        kwargs["Parallel_UseOmpThreads"] = "Yes"
    if dispersion is not None:
        kwargs["Hamiltonian_Dispersion"] = dispersion

    if skf_dir.find("3ob") > 0:
        calc_type = "3ob"
    elif skf_dir.find("mio") > 0:
        calc_type = "mio"
    elif skf_dir.find("pbc") > 0:
        calc_type = "pbc"
    elif skf_dir.find("matsci") > 0:
        calc_type = "matsci"

    # https://dftb.org/parameters/download/3ob/3ob-3-1-cc
    if calc_type == "3ob":
        kwargs["Hamiltonian_ThirdOrderFull"] = "Yes"
        kwargs["Hamiltonian_HCorrection"] = "Damping {\n\tExponent = 4.00\n\t}"
        HD = {
            "Br": -0.0573,
            "C": -0.1492,
            "N": -0.1535,
            "Ca": -0.0340,
            "Na": -0.0454,
            "Cl": -0.0697,
            "Zn": -0.03,
            "O": -0.1575,
            "F": -0.1623,
            "P": -0.14,
            "H": -0.1857,
            "S": -0.11,
            "I": -0.0433,
            "K": -0.0339,
        }
        strs = "{"
        for ele in atom_types:
            if ele == "H":
                kwargs["Hamiltonian_MaxAngularMomentum_H"] = "s"
            elif ele in ["Mg", "C", "N", "Ca", "Na", "O", "F", "K"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "p"
            elif ele in ["Br", "Cl", "P", "S", "I", "Zn"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "d"
            else:
                raise RuntimeError("3-ob-1 doesnot support", ele)
            strs += "\n\t" + ele + " = " + str(HD[ele])
        strs += "\n\t}"
        kwargs["Hamiltonian_HubbardDerivs"] = strs
    elif calc_type == "pbc":
        # https://dftb.org/parameters/download/pbc/pbc-0-3-cc
        for ele in atom_types:
            if ele == "H":
                kwargs["Hamiltonian_MaxAngularMomentum_H"] = "s"
            elif ele in ["C", "O", "N", "F"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "p"
            elif ele in ["Si", "Fe"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "d"
            else:
                raise RuntimeError("pbc-0-3 doesnot support", ele)
    elif calc_type in ["matsci", "mio"]:
        # https://dftb.org/parameters/download/pbc/pbc-0-3-cc
        for ele in atom_types:
            if ele == "H":
                kwargs["Hamiltonian_MaxAngularMomentum_H"] = "s"
            elif ele in ["B", "O", "C", "N"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "p"
            elif ele in ["Si"]:
                kwargs["Hamiltonian_MaxAngularMomentum_" + ele] = "d"
            else:
                raise RuntimeError(calc_type, "doesnot support", ele)

    # DFTB2

    # pbc-0-3
    # matsci
    # ob2
    # pbc
    # print(calc_type, kwargs)
    return kwargs


def DFTB_relax(
    struc,
    skf_dir,
    opt_cell=False,
    step=500,
    fmax=0.1,
    kresol=0.10,
    folder="tmp",
    disp="D3",
    mask=None,
    symmetrize=True,
    logfile=None,
    scc_error=1e-6,
    scc_iter=500,
    use_omp=False,
):
    """
    DFTB optimizer based on ASE

    Args:
        struc: ase atoms object
        mode: [`single`, `relax`, `vc_relax`] (str)
        step: optimization steps (int)
        mask: apply constraints on strain
    """
    if not os.path.exists(folder):
        os.makedirs(folder)
    cwd = os.getcwd()
    os.chdir(folder)
    if type(kresol) != list:
        kpts = Kgrid(struc, kresol)
    atom_types = set(struc.get_chemical_symbols())
    kwargs = make_Hamiltonian(skf_dir, atom_types, disp, kpts, scc_error=scc_error, use_omp=use_omp)

    calc = Dftb(
        label="test",
        atoms=struc,
        kpts=kpts,
        **kwargs,
    )
    struc.set_calculator(calc)

    # impose symmetry
    if symmetrize:
        struc.set_constraint(FixSymmetry(struc))

    # impose cell constraints
    if opt_cell:
        ecf = UnitCellFilter(struc, mask=mask)
        dyn = FIRE(ecf, logfile=logfile)
    else:
        dyn = FIRE(struc, logfile=logfile)

    try:
        dyn.run(fmax=fmax, steps=step)
        os.remove("dftb_pin.hsd")
        os.remove("geo_end.gen")
        os.remove("charges.bin")
    except:
        print("Problem in DFTB calculation")
        struc = None

    os.chdir(cwd)
    return struc


def DFTB_SCF(struc, skf_dir, kresol=0.10, folder="tmp", disp=None, filename=None):
    """
    DFTB SCF to get band structure

    Args:
        struc: ase atoms object
        skf_dir:
        kresol:
        filename: band structure output
    """
    if not os.path.exists(folder):
        os.makedirs(folder)
    cwd = os.getcwd()
    os.chdir(folder)

    kpts = Kgrid(struc, kresol)
    atom_types = set(struc.get_chemical_symbols())
    kwargs = make_Hamiltonian(skf_dir, atom_types, disp, kpts, write_band=True)

    calc = Dftb(
        label="test",
        atoms=struc,
        kpts=kpts,
        **kwargs,
    )

    struc.set_calculator(calc)
    struc.get_potential_energy()
    eigvals = calc.read_eigenvalues()[0]
    ne = calc.read_electrons()
    nband = int(ne / 2)
    vbm = eigvals[:, nband - 1].max()
    cbm = eigvals[:, nband].min()
    gap = cbm - vbm
    # if filename is not None:
    #    # plot band structure
    os.chdir(cwd)
    return gap


class DFTB:
    """
    Modified DFTB calculator

    Args:
        struc: ase atoms object
        disp: dispersion (D3, D30, TS, MBD))
        step: optimization steps (int)
        kresol: kpoint resolution (float)
        folder: (str)
        label: (str)
        use_omp: True/False
    """

    def __init__(
        self,
        struc,
        skf_dir,
        disp="D3",
        kresol=0.10,
        folder="tmp",
        label="test",
        prefix="geo_final",
        use_omp=False,
        scc_error=1e-6,
        scc_iter=500,
    ):
        self.struc = struc
        self.skf_dir = skf_dir
        self.folder = folder
        self.kresol = kresol
        self.disp = disp
        self.label = label
        self.kpts = Kgrid(struc, kresol)
        self.prefix = prefix
        self.use_omp = use_omp
        self.scc_error = scc_error
        self.scc_iter = scc_iter
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

    def get_calculator(self, mode, step=500, ftol=1e-3, FixAngles=False, eVperA=True, md_params=None):
        """
        get the ase style calculator

        Args:
            mode: ['single', 'relax', 'vc_relax'] (str)
            step: relaxation steps (int)
            ftol: force tolerance (float)
            FixAngles: Fix angles between lattice vectors
            eVperA: unit in eV/A

        Returns:
            ase calculator
        """
        if md_params is None:
            md_params = {}
        if eVperA:
            ftol *= 0.194469064593167e-01
        atom_types = set(self.struc.get_chemical_symbols())
        kwargs = make_Hamiltonian(
            self.skf_dir,
            atom_types,
            self.disp,
            self.kpts,
            scc_error=self.scc_error,
            scc_iter=self.scc_iter,
            use_omp=self.use_omp,
        )

        if mode in ["relax", "vc-relax"]:
            # kwargs['Driver_'] = 'ConjugateGradient'
            kwargs["Driver_"] = "GeometryOptimization"
            kwargs["Driver_Optimizer"] = "FIRE {}"
            # kwargs['Driver_MaxForceComponent'] = ftol
            kwargs["Driver_MaxSteps"] = step
            kwargs["Driver_OutputPrefix"] = self.prefix
            kwargs["Driver_Convergence_"] = ""
            kwargs["Driver_Convergence_GradElem"] = ftol

            if mode == "vc-relax":
                kwargs["Driver_MovedAtoms"] = "1:-1"
                kwargs["Driver_LatticeOpt"] = "Yes"
                if FixAngles:
                    kwargs["Driver_FixAngles"] = "Yes"

        elif mode in ["nve", "nvt", "npt"]:
            # 1fs = 41.3 au
            # 1000K = 0.0031668 au
            dicts = {
                "temperature": 300,
                "pressure": 1e5,  # 1atm
                "timestep": 1,
                "Thermostat": "NoseHoover",
                "MDRestartFrequency": 1000,
                "band": "No",
            }
            dicts.update(md_params)

            kwargs["Analysis_WriteBandOut"] = dicts["band"]
            kwargs["Driver_"] = "VelocityVerlet"
            kwargs["Driver_Steps"] = step
            kwargs["Driver_TimeStep [fs]"] = dicts["timestep"]
            kwargs["Driver_MDRestartFrequency"] = dicts["MDRestartFrequency"]
            kwargs["Driver_MovedAtoms"] = "1:-1"
            kwargs["Driver_OutputPrefix"] = self.prefix

            if mode in ["nvt", "npt"]:
                kwargs["Driver_Thermostat_"] = dicts["Thermostat"]
                kwargs["Driver_Thermostat_Temperature [Kelvin]"] = dicts["temperature"]
                kwargs["Driver_Thermostat_CouplingStrength [cm^-1]"] = 3200

                if mode == "npt":
                    kwargs["Driver_Barostat_"] = ""
                    kwargs["Driver_Barostat_Pressure [Pa]"] = dicts["pressure"]
                    kwargs["Driver_Barostat_Timescale [ps]"] = 0.1

        return Dftb(
            label=self.label,
            # run_manyDftb_steps=True,
            atoms=self.struc,
            kpts=self.kpts,
            **kwargs,
        )

    def run(self, mode, step=500, ftol=1e-3, FixAngles=False, md_params=None):
        """
        execute the actual calculation
        """
        from time import time

        if md_params is None:
            md_params = {}
        t0 = time()
        cwd = os.getcwd()
        os.chdir(self.folder)

        self.calc = self.get_calculator(mode, step, ftol, FixAngles, md_params=md_params)
        self.struc.set_calculator(self.calc)
        # self.struc.write('geo_o.gen', format='dftb')
        # execute the simulation
        self.calc.calculate(self.struc)
        final = read(self.prefix + ".gen") if mode in ["relax", "vc-relax"] else self.struc

        # get the final energy
        energy = self.struc.get_potential_energy()

        with open(self.label + ".out") as f:
            l = f.readlines()
            self.version = l[2]
        os.chdir(cwd)
        self.time = time() - t0
        return final, energy


class Dftb(FileIOCalculator):
    """This module defines a FileIOCalculator for DFTB+

    http://www.dftbplus.org/
    http://www.dftb.org/

    Initial development: markus.kaukonen@iki.fi
    Modified by QZ to avoid the I/O load
    """

    command = os.environ["DFTB_COMMAND"] + " > PREFIX.out" if "DFTB_COMMAND" in os.environ else "dftb+ > PREFIX.out"

    implemented_properties = ["energy", "forces", "stress"]
    discard_results_on_any_change = True

    def __init__(
        self,
        restart=None,
        ignore_bad_restart_file=FileIOCalculator._deprecated,
        label="dftb",
        atoms=None,
        kpts=None,
        slako_dir=None,
        **kwargs,
    ):
        """
        All keywords for the dftb_in.hsd input file (see the DFTB+ manual)
        can be set by ASE. Consider the following input file block:

        >>> Hamiltonian = DFTB {
        >>>     SCC = Yes
        >>>     SCCTolerance = 1e-8
        >>>     MaxAngularMomentum = {
        >>>         H = s
        >>>         O = p
        >>>     }
        >>> }

        This can be generated by the DFTB+ calculator by using the
        following settings:

        >>> calc = Dftb(Hamiltonian_='DFTB',  # line is included by default
        >>>             Hamiltonian_SCC='Yes',
        >>>             Hamiltonian_SCCTolerance=1e-8,
        >>>             Hamiltonian_MaxAngularMomentum_='',
        >>>             Hamiltonian_MaxAngularMomentum_H='s',
        >>>             Hamiltonian_MaxAngularMomentum_O='p')

        In addition to keywords specific to DFTB+, also the following keywords
        arguments can be used:

        restart: str
            Prefix for restart file.  May contain a directory.
            Default is None: don't restart.
        ignore_bad_restart_file: bool
            Ignore broken or missing restart file. By default, it is an
            error if the restart file is missing or broken.
        label: str (default 'dftb')
            Prefix used for the main output file (<label>.out).
        atoms: Atoms object (default None)
            Optional Atoms object to which the calculator will be
            attached. When restarting, atoms will get its positions and
            unit-cell updated from file.
        kpts: (default None)
            Brillouin zone sampling:

            * ``(1,1,1)`` or ``None``: Gamma-point only
            * ``(n1,n2,n3)``: Monkhorst-Pack grid
            * ``dict``: Interpreted as a path in the Brillouin zone if
              it contains the 'path_' keyword. Otherwise it is converted
              into a Monkhorst-Pack grid using
              ``ase.calculators.calculator.kpts2sizeandoffsets``
            * ``[(k11,k12,k13),(k21,k22,k23),...]``: Explicit (Nkpts x 3)
              array of k-points in units of the reciprocal lattice vectors
              (each with equal weight)

        """

        if slako_dir is None:
            slako_dir = os.environ.get("DFTB_PREFIX", "./")
            if not slako_dir.endswith("/"):
                slako_dir += "/"

        self.slako_dir = slako_dir

        self.default_parameters = dict(
            Hamiltonian_="DFTB",
            Hamiltonian_SlaterKosterFiles_="Type2FileNames",
            Hamiltonian_SlaterKosterFiles_Prefix=self.slako_dir,
            Hamiltonian_SlaterKosterFiles_Separator='"-"',
            Hamiltonian_SlaterKosterFiles_Suffix='".skf"',
            Hamiltonian_MaxAngularMomentum_="",
            Options_="",
            Options_WriteResultsTag="Yes",
            Options_WriteDetailedOut="No",
        )

        self.lines = None
        self.atoms = None
        self.atoms_input = None
        self.outfilename = "dftb.out"

        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file, label, atoms, **kwargs)

        # kpoint stuff by ase
        self.kpts = kpts
        self.kpts_coord = None

        if self.kpts is not None:
            initkey = "Hamiltonian_KPointsAndWeights"
            mp_mesh = None
            offsets = None

            if isinstance(self.kpts, dict):
                if "path" in self.kpts:
                    # kpts is path in Brillouin zone
                    self.parameters[initkey + "_"] = "Klines "
                    self.kpts_coord = kpts2ndarray(self.kpts, atoms=atoms)
                else:
                    # kpts is (implicit) definition of
                    # Monkhorst-Pack grid
                    self.parameters[initkey + "_"] = "SupercellFolding "
                    mp_mesh, offsets = kpts2sizeandoffsets(atoms=atoms, **self.kpts)
            elif np.array(self.kpts).ndim == 1:
                # kpts is Monkhorst-Pack grid
                self.parameters[initkey + "_"] = "SupercellFolding "
                mp_mesh = self.kpts
                offsets = [0.0] * 3
            elif np.array(self.kpts).ndim == 2:
                # kpts is (N x 3) list/array of k-point coordinates
                # each will be given equal weight
                self.parameters[initkey + "_"] = ""
                self.kpts_coord = np.array(self.kpts)
            else:
                raise ValueError("Illegal kpts definition:" + str(self.kpts))

            if mp_mesh is not None:
                eps = 1e-10
                for i in range(3):
                    key = initkey + "_empty%03d" % i
                    val = [mp_mesh[i] if j == i else 0 for j in range(3)]
                    self.parameters[key] = " ".join(map(str, val))
                    offsets[i] *= mp_mesh[i]
                    assert abs(offsets[i]) < eps or abs(offsets[i] - 0.5) < eps
                    # DFTB+ uses a different offset convention, where
                    # the k-point mesh is already Gamma-centered prior
                    # to the addition of any offsets
                    if mp_mesh[i] % 2 == 0:
                        offsets[i] += 0.5
                key = initkey + "_empty%03d" % 3
                self.parameters[key] = " ".join(map(str, offsets))

            elif self.kpts_coord is not None:
                for i, c in enumerate(self.kpts_coord):
                    key = initkey + "_empty%09d" % i
                    c_str = " ".join(map(str, c))
                    if "Klines" in self.parameters[initkey + "_"]:
                        c_str = "1 " + c_str
                    else:
                        c_str += " 1.0"
                    self.parameters[key] = c_str

    def write_dftb_in(self, outfile):
        """Write the innput file for the dftb+ calculation.
        Geometry is taken always from the file 'geo_end.gen'.
        """

        outfile.write("Geometry = GenFormat { \n")
        outfile.write('    <<< "geo_end.gen" \n')
        outfile.write("} \n")
        outfile.write(" \n")

        params = self.parameters.copy()

        s = "Hamiltonian_MaxAngularMomentum_"
        for key in params:
            if key.startswith(s) and len(key) > len(s):
                break
        # --------MAIN KEYWORDS-------
        previous_key = "dummy_"
        myspace = " "
        for key, value in sorted(params.items()):
            current_depth = key.rstrip("_").count("_")
            previous_depth = previous_key.rstrip("_").count("_")
            for my_backsclash in reversed(range(previous_depth - current_depth)):
                outfile.write(3 * (1 + my_backsclash) * myspace + "} \n")
            outfile.write(3 * current_depth * myspace)
            if key.endswith("_") and len(value) > 0:
                outfile.write(key.rstrip("_").rsplit("_")[-1] + " = " + str(value) + "{ \n")
            elif key.endswith("_") and (len(value) == 0) and current_depth == 0:  # E.g. 'Options {'
                outfile.write(key.rstrip("_").rsplit("_")[-1] + " " + str(value) + "{ \n")
            elif key.endswith("_") and (len(value) == 0) and current_depth > 0:  # E.g. 'Hamiltonian_Max... = {'
                outfile.write(key.rstrip("_").rsplit("_")[-1] + " = " + str(value) + "{ \n")
            elif key.count("_empty") == 1:
                outfile.write(str(value) + " \n")
            else:
                outfile.write(key.rsplit("_")[-1] + " = " + str(value) + " \n")
            previous_key = key
        current_depth = key.rstrip("_").count("_")
        for my_backsclash in reversed(range(current_depth)):
            outfile.write(3 * my_backsclash * myspace + "} \n")
        outfile.write("ParserOptions { \n")
        outfile.write("   IgnoreUnprocessedNodes = Yes  \n")
        outfile.write("} \n")

    def check_state(self, atoms):
        system_changes = FileIOCalculator.check_state(self, atoms)
        # Ignore unit cell for molecules:
        if not atoms.pbc.any() and "cell" in system_changes:
            system_changes.remove("cell")
        return system_changes

    def write_input(self, atoms, properties=None, system_changes=None):
        from ase.io import write

        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        with open(os.path.join(self.directory, "dftb_in.hsd"), "w") as fd:
            self.write_dftb_in(fd)
        write(os.path.join(self.directory, "geo_end.gen"), atoms, parallel=False)
        # self.atoms is none until results are read out,
        # then it is set to the ones at writing input
        self.atoms_input = atoms
        self.atoms = None

    def read_results(self):
        """all results are read from results.tag file
        It will be destroyed after it is read to avoid
        reading it once again after some runtime error"""
        with open(os.path.join(self.directory, "results.tag")) as fd:
            self.lines = fd.readlines()
        if len(self.lines) == 0:
            # print("READ RESULTS from test.out")
            self.results["energy"] = self.read_energy()
            self.results["forces"] = None
            self.results["stress"] = None
        else:
            self.atoms = self.atoms_input
            self.results["energy"] = float(self.lines[1]) * Hartree
            forces = self.read_forces()
            self.results["forces"] = forces

            # stress stuff begins
            sstring = "stress"
            have_stress = False
            stress = list()
            for iline, line in enumerate(self.lines):
                if sstring in line:
                    have_stress = True
                    start = iline + 1
                    end = start + 3
                    for i in range(start, end):
                        cell = [float(x) for x in self.lines[i].split()]
                        stress.append(cell)
            if have_stress:
                stress = -np.array(stress) * Hartree / Bohr**3
                self.results["stress"] = stress.flat[[0, 4, 8, 5, 2, 1]]
            # stress stuff ends

            # calculation was carried out with atoms written in write_input
            os.remove(os.path.join(self.directory, "results.tag"))

    def read_energy(self):
        """
        If SCC is not converged, read the last step energy from test.out
        """
        outfile = self.label + ".out"
        energies = []
        with open(os.path.join(self.directory, outfile)) as fd:
            lines = fd.readlines()
        for line in lines:
            m = re.match(r"Total Energy:\s+[-\d.]+ H\s+([-.\d]+) eV", line)
            if m:
                energies.append(float(m.group(1)))
        if len(energies) > 0:
            return energies[-1]
        else:
            try:
                #   100   -0.38553421E+03    0.29798304E-03    0.12389437E-01
                # ERROR!
                # -> SCC is NOT converged, maximal SCC iterations exceeded
                print("Cannot read energy from this file", os.getcwd())
                return float(lines[-3].split()[1]) * 27.2114  # hatree to eV
            except:
                return 1.0e5

    def read_forces(self):
        """Read Forces from dftb output file (results.tag)."""

        # Initialise the indices so their scope
        # reaches outside of the for loop
        index_force_begin = -1
        index_force_end = -1

        # Force line indexes
        fstring = "forces   "
        for iline, line in enumerate(self.lines):
            if line.find(fstring) >= 0:
                index_force_begin = iline + 1
                line1 = line.replace(":", ",")
                index_force_end = iline + 1 + int(line1.split(",")[-1])
                break
        gradients = []
        for j in range(index_force_begin, index_force_end):
            word = self.lines[j].split()
            gradients.append([float(word[k]) for k in range(3)])
        return np.array(gradients) * Hartree / Bohr

    def read_eigenvalues(self):
        """Read Eigenvalues from dftb output file (results.tag).
        Unfortunately, the order seems to be scrambled."""
        # Eigenvalue line indexes
        index_eig_begin = None
        for iline, line in enumerate(self.lines):
            fstring = "eigenvalues   "
            if line.find(fstring) >= 0:
                index_eig_begin = iline + 1
                line1 = line.replace(":", ",")
                ncol, nband, nkpt, nspin = map(int, line1.split(",")[-4:])
                break
        else:
            return None

        # Take into account that the last row may lack
        # columns if nkpt * nspin * nband % ncol != 0
        nrow = int(np.ceil(nkpt * nspin * nband * 1.0 / ncol))
        index_eig_end = index_eig_begin + nrow
        ncol_last = len(self.lines[index_eig_end - 1].split())
        if ncol - ncol_last > 0:
            self.lines[index_eig_end - 1] = self.lines[index_eig_end - 1].replace("\n", "")
            self.lines[index_eig_end - 1] += " 0.0 " * (ncol - ncol_last)
            self.lines[index_eig_end - 1] += "\n"
        eig = np.loadtxt(self.lines[index_eig_begin:index_eig_end]).flatten()
        eig *= Hartree
        N = nkpt * nband
        return [eig[i * N : (i + 1) * N].reshape((nkpt, nband)) for i in range(nspin)]

    def read_fermi_levels(self):
        """Read Fermi level(s) from dftb output file (results.tag)."""
        # Fermi level line indexes
        for iline, line in enumerate(self.lines):
            fstring = "fermi_level   "
            if line.find(fstring) >= 0:
                index_fermi = iline + 1
                break
        else:
            return None

        fermi_levels = []
        words = self.lines[index_fermi].split()
        assert len(words) in [1, 2], "Expected either 1 or 2 Fermi levels"

        for word in words:
            e = float(word)
            # In non-spin-polarized calculations with DFTB+ v17.1,
            # two Fermi levels are given, with the second one being 0,
            # but we don't want to add that one to the list
            if abs(e) > 1e-8:
                fermi_levels.append(e)

        return np.array(fermi_levels) * Hartree

    def read_electrons(self):
        """read number o electrons"""
        for iline, line in enumerate(self.lines):
            fstring = "number_of_electrons"
            if line.find(fstring) >= 0:
                index_ele = iline + 1
                break
        return float(self.lines[index_ele].split("\n")[0])


if __name__ == "__main__":
    from ase.build import bulk

    skf_dir = os.environ["DFTB_PREFIX"] + "pbc-0-3/"
    # skf_dir = os.environ['DFTB_PREFIX'] + '3ob-3-1/'

    struc = bulk("Si", "diamond", cubic=True)
    struc.set_cell(1.1 * struc.cell)
    for mode in ["single", "relax", "vc-relax", "npt"]:
        my = DFTB(struc, skf_dir)
        struc, energy = my.run(mode)
        res = f"{mode:8s} "
        res += f"{struc.cell[0, 0]:8.4f} "
        res += f"{struc.cell[1, 1]:8.4f} "
        res += f"{struc.cell[2, 2]:8.4f} "
        res += f"{energy:12.4f}"
        print(res)
    # gap = DFTB_SCF(struc, skf_dir)
    # print(gap)


"""
Module to store the constants
"""

import numpy as np
from pyxtal.version import __version__

# Constants
# ------------------------------
tol_m = 0.3  # seperation tolerance in Angstroms
rad = np.pi / 180.0  # converting degrees to radians
deg = 180.0 / np.pi  # converting radians to degrees
pyxtal_verbosity = 1  # constant for printx function
# Max number of atoms per molecule before using fast distance check
max_fast_mol_size = 30
letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
ltype_keywords = [
    "triclinic",
    "Triclinic",
    "monoclinic",
    "Monoclinic",
    "orthorhombic",
    "Orthorhombic",
    "tetragonal",
    "Tetragonal",
    "trigonal",
    "Trigonal",
    "hexagonal",
    "Hexagonal",
    "cubic",
    "Cubic",
    "spherical",
    "Spherical",
    "ellipsoidal",
    "Ellipsoidal",
]
single_smiles = [
    "Cl-",
    "F-",
    "Br-",
    "I-",
    "Li+",
    "Na+",
    "Cs+",
    "Rb+",
    "[Cl-]",
    "[F-]",
    "[Br-]",
    "[I-]",
    "[Li+]",
    "[Na+]",
    "[Cs+]",
    "Rb+",
]
hex_cell = np.array([[1, -0.5, 0], [0, np.sqrt(3) / 2, 0], [0, 0, 1]])
all_sym_directions = [
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1),
    (1, 1, 1),
    (1, -1, -1),
    (-1, 1, -1),
    (-1, -1, 1),
    (1, -1, 0),
    (1, 1, 0),
    (0, 1, -1),
    (0, 1, 1),
    (-1, 0, 1),
    (1, 0, 1),
    (1, -2, 0),
    (2, -1, 0),
]

logo = rf"""#############################################################
#             ______       _    _          _                #
#            (_____ \     \ \  / /        | |               #
#             _____) )   _ \ \/ / |_  ____| |               #
#            |  ____/ | | | )  (|  _)/ _  | |               #
#            | |    | |_| |/ /\ \ |_( (_| | |___            #
#            |_|     \__  /_/  \_\___)__|_|_____)           #
#                   (____/                                  #
#---------------------(version {__version__:>8s})--------------------#
#       A Python package for random crystal generation      #
#       url: https://github.com/qzhu2017/pyxtal             #
#       @Zhu's group at University of Nevada Las Vegas      #
#############################################################
"""



"""
Module for handling Wyckoff sites for both atom and molecule
"""

# Standard Libraries
from copy import deepcopy
import numpy as np

# External Libraries
from pymatgen.core import Molecule
from scipy.spatial.distance import cdist
from scipy.spatial.transform import Rotation as R

from pyxtal.constants import rad
from pyxtal.database.element import Element
from pyxtal.lattice import Lattice
from pyxtal.operations import (
    SymmOp,
    create_matrix,
    distance_matrix,
    filtered_coords,
)
from pyxtal.symmetry import Group, Wyckoff_position
from pyxtal.tolerance import Tol_matrix


class atom_site:
    """
    Class for storing atomic Wyckoff positions with a single coordinate.

    Args:
        wp: a `WP <pyxtal.symmetry.Wyckoff_position.html>` object
        coordinate (float): a fractional (x, y, z) coordinate
        specie (str): element name or symbol, or atomic number
        search (bool): whether or not search generator for special WP
    """

    def __init__(self, wp=None, coordinate=None, specie=1, search=False):
        self.position = np.array(coordinate)
        self.position -= np.floor(self.position)
        self.specie = Element(specie).short_name
        self.wp = wp
        self.coordination = None

        self._get_dof()
        self.PBC = self.wp.PBC
        self.multiplicity = self.wp.multiplicity
        if search:
            self.search_position()

        self.update()

    def __str__(self):
        if not hasattr(self.wp, "site_symm"):
            self.wp.get_site_symmetry()
        s = "{:>2s} @ [{:7.4f} {:7.4f} {:7.4f}], ".format(
            self.specie, *self.position)
        s += f"WP [{self.wp.get_label()}] "
        if self.coordination is not None:
            s += f" CN [{self.coordination:2d}] "
        s += "Site [{:}]".format(self.wp.site_symm.replace(" ", ""))

        return s

    def __repr__(self):
        return str(self)

    def copy(self):
        """
        Simply copy the structure
        """

        return deepcopy(self)

    def save_dict(self):
        return {
            "position": self.position,
            "specie": self.specie,
            "wp": self.wp.save_dict(),
        }

    def _get_dof(self):
        """
        get the number of dof for the given structures:
        """

        self.dof = self.wp.get_dof()

    def get_bounds(self):
        """
        get the number of dof for the given structures:
        """
        self.bounds = []
        for i in range(self.dof):
            self.bounds.append([0.0, 1.0])
        return self.bounds

    @classmethod
    def load_dict(cls, dicts):
        """
        load the sites from a dictionary
        """

        position = dicts["position"]
        specie = dicts["specie"]
        if "wp" in dicts:
            wp = Wyckoff_position.load_dict(dicts["wp"])
        else:
            hn, index = dicts["hn"], dicts["index"]
            wp = Wyckoff_position.from_group_and_index(
                hn, index, use_hall=True)

        return cls(wp, position, specie)

    def perturbate(self, lattice, magnitude=0.1):
        """
        Random perturbation of the site

        Args:
            lattice: lattice vectors
            magnitude: the magnitude of displacement (default: 0.1 A)
        """
        dis = (np.random.sample(3) - 0.5).dot(lattice)
        dis /= np.linalg.norm(dis)
        dis *= magnitude
        pos = self.position + dis.dot(np.linalg.inv(lattice))
        self.update(pos)

    def search_position(self, tol=1e-3):
        """
        Sometimes, the initial position is not the proper generator
        Needs to find the proper generator
        """

        found = False
        if self.wp.index > 0:
            wp0 = Group(self.wp.number, self.wp.dim)[0]
            for coord in wp0.apply_ops(self.position):
                coord -= np.floor(coord)
                ans = self.wp.ops[0].operate(coord)
                diff = coord - ans
                diff -= np.rint(diff)
                if np.sum(diff**2) < tol:
                    found = True
                    # print(found, coord, coord-ans)
                    self.position = coord - np.floor(coord)
                    break

        if not found:
            print("\nInput xyz", self.position)
            print("Target operation", self.wp.ops[0].as_xyz_str())
            raise ValueError("Cannot generate the desried generator")

    def encode(self):
        """
        transform dict to 1D vector
        [specie, wp.index, free x, y, z]
        """
        xyz = self.wp.get_free_xyzs(self.position)
        # print(self.wp.ops[0].rotation_matrix, self.wp.get_frozen_axis(), self.wp.get_dof())
        # print([self.specie, self.wp.index] + list(xyz))
        return [self.specie, self.wp.index, *list(xyz)]


    def shift_by_swap(self, swap_id):
        """
        Check if a shift is needed during swap.

        This may occur for special WP in the I/A/B/C/F cases.
        For example, in space group 71 (Immm), the permutation:
        4j(1/2,0,0.2) -> (0.2,0,1/2) -> 4f(0.7,1/2,0) 
        requires a shift of (0.5,0.5,0.5)

        Args:
            swap_id: The index of the swap operation

        Returns:
            numpy array: The required shift vector
        """
        wp, shift = self.wp.swap_axis(swap_id)
        return shift

    def equivalent_set(self, tran, indices):
        """
        Transform the wp to another equivalent set.
        Needs to update both wp and positions

        Args:
            tran: affine matrix
            indices: the list of transformed wps
        """
        self.position = SymmOp(tran).operate(self.position)
        self.position -= np.floor(self.position)
        self.wp = self.wp.equivalent_set(
            indices[self.wp.index])  # update the wp index
        self.site_symm = self.wp.site_symm  # update the site symmetry
        self.update()

    def update(self, pos=None, reset_wp=False):
        """
        Used to generate coords from self.position
        """

        if pos is None:
            pos = self.position
        if reset_wp:
            self.wp.ops = Group(self.wp.number)[self.wp.index].ops
        self.coords = self.wp.apply_ops(pos)
        self.position = self.coords[0]

    def get_translations(self, pos, axis):
        """
        Get the displacement towards the reference positions

        Args:
            pos: reference position (1*3 vector)
            lattice: 3*3 matrix
            translation:
            axis:
        """

        # diffs0 = pos - self.coords
        diffs0 = self.wp.apply_ops(pos) - self.position
        diffs = diffs0.copy()
        diffs -= np.rint(diffs)
        diffs[:, axis] = 0
        return diffs0 - diffs

    def get_disp(self, pos, lattice, translation):
        """
        return the displacement towards the reference positions

        Args:
            pos: reference position (1*3 vector)
            lattice: 3*3 matrix
            translation:
        """
        coords = self.wp.apply_ops(pos)
        diffs = coords - (self.position + translation)
        # coords = self.wp.apply_ops(self.position + translation)
        # diffs = pos - coords

        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs.dot(lattice), axis=1)
        id = np.argmin(dists)

        # print("++++", id, dists[id], id, diffs[id], translation) #; import sys; sys.exit()
        return diffs[id], dists[id]

    def check_with_ws2(self, ws2, lattice, tm, same_group=True):
        """
        Given two Wyckoff sites, checks the inter-atomic distances between them.

        Args:
            - ws2: a different WP object (will always return False if
            two identical WS's are provided)
            - lattice: a 3x3 cell matrix
            - same_group: whether or not two WS's are in the same structure.
            Default value True reduces the calculation cost
        Returns:
            True or False
        """

        # Ensure the PBC values are valid
        if self.PBC != ws2.PBC:
            raise ValueError("PBC values do not match between Wyckoff sites")

        # Get tolerance
        tol = tm.get_tol(self.specie, ws2.specie)
        # Symmetry shortcut method: check only some atoms
        if same_group is True:
            # We can either check one atom in WS1 against all WS2, or vice-versa
            # Check which option is faster
            if self.multiplicity > ws2.multiplicity:
                coords1 = [self.coords[0]]
                coords2 = ws2.coords
            else:
                coords1 = [ws2.coords[0]]
                coords2 = self.coords
            # Calculate distances
            dm = distance_matrix(coords1, coords2, lattice, PBC=self.PBC)
            # Check if any distances are less than the tolerance
            return not (dm < tol).any()
        # No symmetry method: check all atomic pairs
        else:
            dm = distance_matrix(
                self.wp.coords, ws2.coords, lattice, PBC=self.PBC)
            # Check if any distances are less than the tolerance
            return not (dm < tol).any()

    def substitute_with_single(self, ele):
        """
        chemical substitution with another element

        Args:
            ele (str): e.g. 'Zn'
        """
        self.specie = ele
        return self

    def substitute_with_linear(self, eles, direction, lattice):
        """
        chemical substitution with another linear building block, e.g. CN

        Args:
            eles (str): e.g., ['C', 'N']
            neighbors: two nearest neiboring atom xyz
        """
        bond_length = 1.4
        # direction = neighbors[1] - neighbors[0]
        # direction = np.dot(direction, lattice)
        # direction /= np.linalg.norm(direction)

        # Get the fraction coordinates
        shift = np.dot(bond_length / 2 * direction, np.linalg.inv(lattice))
        site1 = self.copy()
        site2 = self.copy()
        site1.specie = eles[0]
        site2.specie = eles[1]
        site1.update(site1.position + shift)
        site2.update(site2.position - shift)
        return site1, site2

    def to_mol_site(self, lattice, molecule, ori=None, reflect=False, type_id=0):
        """
        Transform it to the mol_sites, i.e., to build a molecule on
        the current WP
        """

        if ori is None:
            ori = [0, 0, 0]
        dicts = {}
        dicts["smile"] = molecule.smile
        dicts["type"] = type_id
        dicts["dim"] = 3
        dicts["PBC"] = [1, 1, 1]
        dicts["hn"] = self.wp.hall_number
        dicts["index"] = self.wp.index
        dicts["lattice"] = lattice.matrix
        dicts["lattice_type"] = lattice.ltype
        dicts["center"] = self.position
        if molecule.smile not in ["Cl-"]:
            dicts["orientation"] = np.array(ori)
            dicts["rotor"] = molecule.get_torsion_angles()
            dicts["reflect"] = reflect

        return mol_site.from_1D_dicts(dicts)


class mol_site:
    """
    Class for storing molecular Wyckoff positions and orientations within
    the `molecular_crystal` class. Each `mol_site` object represents an
    entire Wyckoff position, not necessarily a single molecule.
    This is the molecular version of the `Wyckoff_site` class.

    Attributes:
        mol (pyxtal_molecule): A pyxtal_molecule object representing the molecule at the site.
        position (list or array): A 3-vector representing the generating molecule's position
                                  in fractional coordinates.
        orientation (Orientation): An orientation object that describes the molecule's orientation.
        wp (Wyckoff_position): A Wyckoff position object that holds symmetry information.
        lattice (Lattice): A lattice object that defines the crystal structure's unit cell.
        stype (int): An integer specifying the type of molecule. Default is 0.
        symbols (list): List of atomic symbols for the atoms in the molecule.
        numbers (list): List of atomic numbers for the atoms in the molecule.
        PBC (list): Periodic boundary conditions inherited from the Wyckoff position.
        radius (float): Radius of the molecule, typically used for collision detection.
        tols_matrix (numpy array): Tolerance matrix for the molecular structure.

    Args:
        mol (pyxtal_molecule): A `pyxtal_molecule` object that describes the molecule.
        position (list or array): The 3D fractional coordinates of mol_center in the unit cell.
        orientation (Orientation): The orientation object describing the molecule's rotation.
        wp (Wyckoff_position): A `Wyckoff_position` object defining the symmetry of the site.
        lattice (Lattice, optional): The lattice of the crystal. Can be either a Lattice object
                                     or a matrix that will be converted into a Lattice object.
        stype (int, optional): Integer specifying the type of molecule. Default is 0.

    Methods:
        _get_dof(): Internal method to calculate the degrees of freedom (DoF) for the molecule.
    """

    def __init__(self, mol, position, orientation, wp, lattice=None, stype=0):
        # describe the molecule
        self.molecule = mol
        self.wp = wp
        self.position = position  # fractional coordinate of molecular center
        self.orientation = orientation  # pyxtal.molecule.orientation object
        self._get_dof()
        if isinstance(lattice, Lattice):
            self.lattice = lattice
        else:
            self.lattice = Lattice.from_matrix(lattice)
        self.PBC = self.wp.PBC
        # self.mol = mol.mol # A Pymatgen molecule object
        # [site.specie.value for site in self.mol.sites]
        self.symbols = mol.symbols
        self.numbers = self.molecule.mol.atomic_numbers
        self.tols_matrix = mol.tols_matrix
        self.radius = mol.radius
        self.type = stype

    def update_molecule(self, mol):
        self.molecule = mol
        self.numbers = mol.mol.atomic_numbers
        self.symbols = mol.symbols
        self.tols_matrix = mol.tols_matrix
        self.radius = mol.radius

    def update_orientation(self, angles):
        # QZ: Symmetrize the angle to the compatible orientation first
        self.orientation.r = R.from_euler("zxy", angles, degrees=True)
        self.orientation.matrix = self.orientation.r.as_matrix()

    def get_min_dist(self, angle=None):
        """
        Compute the minimum interatomic distance within the WP.

        Returns:
            minimum distance
        """
        if angle is not None:
            matrix = self.orientation.change_orientation(angle, update=False)
        else:
            matrix = self.orientation.matrix

        # Get total distances
        d0, _ = self.get_dists_auto(matrix=matrix)
        d1, _ = self.get_dists_WP(matrix=matrix, ignore=True)
        d_total = np.append(d0, d1, axis=0)

        return d_total.min()

    def optimize_orientation_by_dist(self, ori_attempts=10, verbose=False):
        """
        Optimize the orientation based on the shortest distance
        """
        # Set initial fun value and angle bounds
        ang_lo, ang_hi = 0, np.pi
        fun_lo = self.get_min_dist(ang_lo)
        fun_hi = self.get_min_dist(ang_hi)
        fun = fun_hi

        # Refine the orientation using a bisection method
        for _it in range(ori_attempts):

            # Return as soon as a good orientation is found
            if (fun > 0.8) and not self.short_dist():
                break

            # Compute the midpoint angle for bisection
            ang = (ang_lo + ang_hi) / 2
            fun = self.get_min_dist(ang)

            # Update based on the function value at the midpoint
            if fun_lo > fun_hi:
                ang_hi, fun_hi = ang, fun
            else:
                ang_lo, fun_lo = ang, fun

            if verbose: print("optimize_orientation_by_dist", _it, ang, fun)

        # Final adjustment
        if fun > fun_hi + 1e-4:
            self.orientation.change_orientation(ang_hi, update=True)
            fun = fun_hi
        elif fun > fun_lo + 1e-4:
            self.orientation.change_orientation(ang_lo, update=True)
            fun = fun_lo
        else:
            self.orientation.change_orientation(ang, update=True)

        return fun

    def get_energy(self, angle=None, wps=[], k1=1.0, r1=1.0, k2=2.0, k3=1.0, r2=6.0,
                   req=2.8, r_cut=2.0, verbose=False):
        """
        Compute the virtual energy based on Hbonding and repulsion

        Args:
            angle (float): rotation along a rotation axis
            wps (list): list of wps for consideration
        """
        # Set the orientation and compute the distance
        if angle is not None:
            matrix = self.orientation.change_orientation(angle, update=False)
        else:
            matrix = self.orientation.matrix

        coord_ref, _ = self._get_coords_and_species(first=True, unitcell=True, matrix=matrix)
        coord_ref = coord_ref @ self.lattice.matrix

        # Get total distances
        eng = 0
        d0, ids_0 = self.get_dists_auto(matrix=matrix, cutoff=8.0)
        d1, ids_1 = self.get_dists_WP(matrix=matrix, ignore=True, cutoff=8.0)
        d_total = np.append(d0, d1, axis=0)
        ids_total = np.append(ids_0, ids_1, axis=0)

        # initialize ids dicts
        acceptor_ids = {}; acceptor_ids[self.type] = self.molecule.active_sites[0]
        donor_ids = {}; donor_ids[self.type] = self.molecule.active_sites[1]
        H_ids = {}; H_ids[self.type] = self.molecule.active_sites[2]

        # Extract short distances ([x, y, z, d], [mol1, mol2, atom_id1, atom_id2])
        for wp2 in wps:
            d2, ids_2 = self.get_dists_WP2(wp2, matrix=matrix, ignore=True, cutoff=8.0)
            if len(d2) > 0:
                d_total = np.append(d_total, d2, axis=0)
                ids_total = np.append(ids_total, ids_2, axis=0)
                if wp2.type not in donor_ids.keys():
                    acceptor_ids[wp2.type] = wp2.molecule.active_sites[0]
                    donor_ids[wp2.type] = wp2.molecule.active_sites[1]
                    H_ids[wp2.type] = wp2.molecule.active_sites[2]

        # Count only 1 contribution per acceptor
        for id1 in self.molecule.active_sites[0]:
            rA = coord_ref[id1]
            dAD, angle, dAH = self.get_dist_angle_AD(d_total, ids_total, id1, donor_ids, H_ids, rA)
            eng += k2 * (dAD-req) ** 2
            eng += k3 * (angle-np.pi) ** 2
            #print(dAD, dAH, angle)
            if dAH < r_cut: eng -= k1 * dAH * np.exp(r_cut - dAH) - 1.0
            if verbose:
                print('Hbond AD', id1, dAD, dAH, np.degrees(angle), eng)

        # Count only 1 contribution per donor
        for i, id1 in enumerate(self.molecule.active_sites[1]):
            rD = coord_ref[id1]
            rH = coord_ref[self.molecule.active_sites[2][i]]

            dAD, angle, dAH = self.get_dist_angle_DA(d_total, ids_total, id1, acceptor_ids, rD, rH)
            eng += k2 * (dAD-req) ** 2
            eng += k3 * (angle-np.pi) ** 2
            if dAH < r_cut: eng -= k1 * dAH * np.exp(r_cut - dAH) - 1.0
            #print(dAD, dAH, angle)
            if verbose:
                print('Hbond DA', id1, dAD, dAH, np.degrees(angle), eng)

        # Count repulsion
        ds = d_total[:, 3]
        ds = ds[ds < r_cut]
        eng += k1 * (ds * np.exp(r_cut - ds) - 1.0).sum()
        #print(ds)
        if verbose:
            print('Repulsion', d_total[d_total < r_cut], eng)

        return eng

    def get_dist_angle_AD(self, d_total, ids_total, A_id, D_ids, H_ids, rA):
        """
        Calculates distances/angles for H bonds between acceptor (A), donor (D), and hydrogen (H) atoms.

        Args:
            d_total (numpy.ndarray): Array containing atomic coordinates and distances
            ids_total (numpy.ndarray): Array containing atom indices and molecule IDs
            A_id (int): Index of the acceptor atom
            D_ids (dict): Dictionary mapping molecule IDs to donor atom indices
            H_ids (dict): Dictionary mapping molecule IDs to hydrogen atom indices
            rA (numpy.ndarray): Coordinates of the acceptor atom

        Returns:
            tuple: A tuple containing three float values:
            - dAD (float): Distance between acceptor and donor atoms
            - angle (float): Angle between H-D and H-A vectors in radians
            - dHA (float): Distance between hydrogen and acceptor atoms
            If no donor atoms are found, returns (10.0, 0, 10.0)
        """

        # Get satisfied
        rows0 = np.where((ids_total[:, 2]==A_id))[0]
        rows_D = []
        rows_H = []
        for row in rows0:
            m2 = ids_total[row, 1]
            if ids_total[row, 3] in D_ids[m2]:
                rows_D.append(row)
            elif ids_total[row, 3] in H_ids[m2]:
                rows_H.append(row)

        if len(rows_D) == 0:
            return 10.0, 0, 10.0
        else:
            myid = d_total[rows_D][:, 3].argmin()
            rD = d_total[rows_D][myid, :3]
            dAD = d_total[rows_D][myid, 3]

            # Find the shortest distance from H
            if len(rows_H) == 0:
                rH = rD + np.array([1, 0, 0])
            else:
                coords_H = d_total[rows_H]
                myid_H = coords_H[:, 3].argmin()
                rH = coords_H[myid_H][:3]

            rHD = rD - rH
            rHA = rA - rH
            dHD = np.linalg.norm(rHD)
            dHA = np.linalg.norm(rHA)
            if abs(dAD - np.linalg.norm(rD-rA)) > 1e-3:
                print("bug", dAD, rD, rA, np.linalg.norm(rD-rA))
                import sys; sys.exit()
            #print('debug dHA, dHD', dHA, dHD)
            cos = rHD @ rHA / (dHD * dHA)
            angle = np.arccos(np.clip(cos, -1.0, 1.0))

            return dAD, angle, dHA

    def get_dist_angle_DA(self, d_total, ids_total, D_id, A_ids, rD, rH):

        # Get satisfied
        rows0 = np.where((ids_total[:, 2]==D_id))[0]
        rows_A = []
        for row in rows0:
            m2 = ids_total[row, 1]
            if ids_total[row, 3] in A_ids[m2]:
                rows_A.append(row)

        if len(rows_A) == 0:
            return 10.0, 0, 10.0
        else:
            myid = d_total[rows_A][:, 3].argmin()
            rA = d_total[rows_A][myid, :3]
            dAD = d_total[rows_A][myid, 3]
            rHD = rD - rH
            rHA = rA - rH
            dHA = np.linalg.norm(rHA)
            dHD = np.linalg.norm(rHD)
            d_min = d_total[myid, 3]
            if abs(dAD - np.linalg.norm(rD-rA)) > 1e-3:
                print("bug", dAD, rD, rA, np.linalg.norm(rD-rA))
                import sys; sys.exit()
            #print('debug dHA, dHD', dHA, dHD)
            cos = rHD @ rHA / (dHD * dHA)
            angle = np.arccos(np.clip(cos, -1.0, 1.0))

            return dAD, angle, dHA

    def optimize_orientation_by_energy(self, wps=[], max_ax=20, max_ori=5, early_quit=3.0, verbose=False):
        """
        Iteratively optimize the orientation with the bisection method
        """
        for ax_trial in range(max_ax):

            # Select axis and compute the initial fun value and angle bounds
            # Select perpendicular????
            self.orientation.set_axis()
            ang_lo = 0 #self.orientation.angle
            ang_hi = np.pi #ang_lo + np.pi
            fun_lo = self.get_energy(wps=wps) #; print("call funlo", fun_lo)
            fun_hi = self.get_energy(ang_hi, wps=wps) #; print("call funhi", fun_hi)
            fun = fun_hi
            #if verbose: print("Init", ang_lo, fun_lo)

            # Refine the orientation using a bisection method
            for ori_trial in range(max_ori):

                # Compute the midpoint angle for bisection
                ang = (ang_lo + ang_hi) / 2
                fun = self.get_energy(ang, wps=wps)
                # Update based on the function value at the midpoint
                if fun_lo < fun_hi:
                    ang_hi, fun_hi = ang, fun
                else:
                    ang_lo, fun_lo = ang, fun
                #print("debug", ang_lo, ang_hi, fun_lo, fun_hi)
            # Finally pick the best one adjustment
            if fun > fun_hi + 1e-4:
                self.orientation.change_orientation(ang_hi, update=True)
                fun = fun_hi
            elif fun > fun_lo + 1e-4:
                self.orientation.change_orientation(ang_lo, update=True)
                fun = fun_lo
            else:
                self.orientation.change_orientation(ang, update=True)

            if verbose: print(f'Final {ax_trial:2d} {fun:.2f}')

            if fun <= early_quit:
                break

    def cut_lattice(self, ax, cut):
        """
        Cut lattice length on the given direction

        Args:
            ax (int): 0, 1, 2
            cut (float): the cut
        """
        paras = self.lattice.get_para()
        x0 = self.position[ax]
        x0 -= np.floor(x0)

        if x0 < 0.25:
            self.position[ax] = paras[ax] * x0 / (paras[ax]-cut)
        elif 0.25 <= x0 <= 0.75:
            self.position[ax] = (paras[ax] * x0 - 0.5 * cut) / (paras[ax]-cut)
        else:
            self.position[ax] = (paras[ax] * x0 - cut) / (paras[ax]-cut)
        #self.lattice.update_para(ax, -cut)

    def __str__(self):
        if not hasattr(self.wp, "site_symm"):
            self.wp.get_site_symmetry()

        self.angles = self.orientation.r.as_euler("zxy", degrees=True)
        formula = self.molecule.mol.formula.replace(" ", "")
        s = "{:12s} @ [{:7.4f} {:7.4f} {:7.4f}]  ".format(
            formula, *self.position)
        s += f"WP [{self.wp.get_label():s}] "
        s += "Site [{:}]".format(self.wp.site_symm.replace(" ", ""))
        if len(self.molecule.mol) > 1:
            s += " Euler [{:6.1f} {:6.1f} {:6.1f}]".format(*self.angles)

        return s

    def __repr__(self):
        return str(self)

    def _get_dof(self):
        """
        get the number of dof for the given wyckoff site:
        """
        dof = np.linalg.matrix_rank(self.wp.ops[0].rotation_matrix)
        self.dof = dof + 3
        if self.molecule.torsionlist is not None:
            self.dof += len(self.molecule.torsionlist)

    def get_bounds(self):
        """
        get the number of dof for the given structures:
        """
        self.bounds = []
        dof = np.linalg.matrix_rank(self.wp.ops[0].rotation_matrix)
        for i in range(dof):
            self.bounds.append([0.0, 1.0])
        self.bounds.append([-180.0, 180.0])
        self.bounds.append([-90.0, 90.0])
        self.bounds.append([-180.0, 180.0])
        if self.molecule.torsionlist is not None:
            for i in range(len(self.molecule.torsionlist)):
                self.bounds.append([0, 180.0])
        return self.bounds

    def save_dict(self):
        dict0 = {
            "position": self.position,
            "wp": self.wp.save_dict(),
            "molecule": self.molecule.save_str(),
            "orientation": self.orientation.save_dict(),
            "lattice": self.lattice.matrix,
            "lattice_type": self.lattice.ltype,
            "stype": self.type,
        }
        if self.molecule.torsionlist is not None:
            xyz, _ = self._get_coords_and_species(absolute=True, first=True)
            dict0["rotors"] = self.molecule.get_torsion_angles(xyz)

        return dict0

    @classmethod
    def load_dict(cls, dicts):
        """
        load the sites from a dictionary
        """
        from pyxtal.molecule import Orientation, pyxtal_molecule

        mol = pyxtal_molecule.load_str(dicts["molecule"])
        position = dicts["position"]
        orientation = Orientation.load_dict(dicts["orientation"])
        wp = Wyckoff_position.load_dict(dicts["wp"])
        lattice = Lattice.from_matrix(
            dicts["lattice"], ltype=dicts["lattice_type"])
        stype = dicts["stype"]
        return cls(mol, position, orientation, wp, lattice, stype)

    def encode(self):
        """
        transform dict to 1D vector
        [wp_id, x, y, z, or1, or2, or3, rotor1, rotor2, .etc]
        """
        if len(self.molecule.mol) > 1:
            xyz, _ = self._get_coords_and_species(absolute=True, first=True)
            # if len(xyz)==3: print("encode: \n", self.molecule.mol.cart_coords)
            rotor = self.molecule.get_torsion_angles(xyz)
            ori, _, reflect = self.molecule.get_orientation(xyz)
            return [self.wp.index, *list(self.position), *list(ori), *rotor, reflect]
        else:
            return [self.wp.index, *list(self.position), 0]

    def to_1D_dicts(self):
        """
        save the wp in 1D representation
        """
        xyz, _ = self._get_coords_and_species(absolute=True, first=True)
        dict0 = {"smile": self.molecule.smile}
        dict0["rotor"] = self.molecule.get_torsion_angles(xyz)
        dict0["orientation"], dict0["rmsd"], dict0["reflect"] = self.molecule.get_orientation(
            xyz)
        dict0["rotor"]
        # rdkit_mol = self.molecule.rdkit_mol(self.molecule.smile)
        # conf0 = rdkit_mol.GetConformer(0)
        # print(self.molecule.set_torsion_angles(conf0, angs))
        # print("save matrix"); print(self.orientation.r.as_matrix())
        # print("save angle"); print(self.orientation.r.as_euler('zxy', degrees=True))
        # print("angle"); print(dict0["orientation"])
        # self.molecule.get_center(xyz)
        dict0["center"] = self.position - np.floor(self.position)
        dict0["number"] = self.wp.number
        dict0["index"] = self.wp.index
        dict0["PBC"] = self.wp.PBC
        dict0["dim"] = self.wp.dim
        dict0["lattice"] = self.lattice.matrix
        dict0["lattice_type"] = self.lattice.ltype

        return dict0

    @classmethod
    def from_1D_dicts(cls, dicts):
        from pyxtal.molecule import Orientation, pyxtal_molecule

        mol = pyxtal_molecule(mol=dicts["smile"] + ".smi", fix=True)
        if len(mol.mol) > 1:
            if len(dicts["smile"]) > 1:
                conf = mol.rdkit_mol().GetConformer(0)
                if dicts["reflect"]:
                    mol.set_torsion_angles(conf, dicts["rotor"], False)
                xyz = mol.set_torsion_angles(
                    conf, dicts["rotor"], dicts["reflect"])
            else:
                # for H2O, use the standard one
                xyz = np.array(
                    [
                        [-0.00111384, 0.36313718, 0.0],
                        [-0.82498189, -0.18196256, 0.0],
                        [0.82609573, -0.18117463, 0.0],
                    ]
                )

            mol.reset_positions(xyz)
            matrix = R.from_euler(
                "zxy", dicts["orientation"], degrees=True).as_matrix()
            orientation = Orientation(matrix)
        else:
            orientation = Orientation(np.eye(3))

        g = dicts["hn"]
        index = int(dicts["index"])
        dim = dicts["dim"]
        wp = Wyckoff_position.from_group_and_index(g, index, dim, dicts["PBC"])
        lattice = Lattice.from_matrix(
            dicts["lattice"], ltype=dicts["lattice_type"])
        # np.dot(dicts["center"], lattice.inv_matrix)
        position = dicts["center"]
        position, wp, _ = wp.merge(position, lattice.matrix, 0.01)

        return cls(mol, position, orientation, wp, lattice)

    def show(self, id=None, **kwargs):
        """
        display WP on the notebook
        """
        from pyxtal.viz import display_molecular_site

        return display_molecular_site(self, id, **kwargs)

    def _get_coords_and_species(self, absolute=False, PBC=False,
                                first=False, unitcell=False,
                                matrix=None):
        """
        Used to generate coords and species for get_coords_and_species

        Args:
            absolute: return absolute or relative coordinates
            PBC: whether or not to add coordinates in neighboring unit cells,
            first: whether or not to extract the information from only the first site
            unitcell: whether or not to move the molecular center to the unit cell
            matrix: orientatin matrix

        Returns:
            atomic coords: a numpy array of atomic coordinates in the site
            species: a list of atomic species for the atomic coords
        """
        if matrix is None: matrix = self.orientation.matrix
        coord0 = self.molecule.mol.cart_coords.dot(matrix.T)
        wp_atomic_sites = []
        wp_atomic_coords = None

        for point_index, op2 in enumerate(self.wp.ops):
            # Obtain the center in absolute coords
            center_relative = op2.operate(self.position)
            if unitcell:
                center_relative -= np.floor(center_relative)
            center_absolute = np.dot(center_relative, self.lattice.matrix)

            # Rotate the molecule (Euclidean metric)
            # op2_m = self.wp.generators_m[point_index]
            op2_m = self.wp.get_euclidean_generator(
                self.lattice.matrix, point_index)
            rot = op2_m.affine_matrix[:3, :3].T
            # NOTE=====the euclidean_generator has wrong translation vectors,
            # but we don't care. This needs to be fixed later
            tmp = np.dot(coord0, rot)  # + tau

            # Add absolute center to molecule
            tmp += center_absolute
            tmp = tmp.dot(self.lattice.inv_matrix)
            wp_atomic_coords = tmp if wp_atomic_coords is None else np.append(
                wp_atomic_coords, tmp, axis=0)
            wp_atomic_sites.extend(self.symbols)
            if first:
                break

        if PBC:
            # Filter PBC of wp_atomic_coords
            wp_atomic_coords = filtered_coords(wp_atomic_coords, PBC=self.PBC)
            # Add PBC copies of coords
            m = create_matrix(PBC=self.PBC, omit=True)
            # Move [0,0,0] PBC vector to first position in array
            m2 = [[0, 0, 0], *list(m)]
            new_coords = np.vstack([wp_atomic_coords + v for v in m2])
            wp_atomic_coords = new_coords

        if absolute:
            wp_atomic_coords = wp_atomic_coords.dot(self.lattice.matrix)

        return wp_atomic_coords, wp_atomic_sites

    def get_coords_and_species(self, absolute=False, PBC=False, unitcell=False):
        """
        Lazily generates and returns the atomic coordinate and species for the
        Wyckoff position. Plugs the molecule into the provided orientation
        (with angle=0), and calculates the new positions.

        Args:
            absolute: return absolute or relative coordinates
            PBC: whether or not to add coordinates in neighboring unit cells
            unitcell: whether or not to move the molecule center to the unit cell

        Returns:
            coords: a np array of 3-vectors.
            species: a list of atomic symbols, e.g. ['H', 'H', 'O', 'H', 'H', 'O']
        """
        return self._get_coords_and_species(absolute, PBC, unitcell=unitcell)

    def perturbate(self, lattice, trans=0.1, rot=5):
        """
        Random perturbation of the molecular site

        Args:
            lattice: lattice vectors
            trans: magnitude of tranlation vectors (default: 0.1 A)
            rot: magnitude of rotation degree (default: 5.0)
        """
        dis = (np.random.sample(3) - 0.5).dot(lattice)
        dis /= np.linalg.norm(dis)
        dis *= trans
        self.translate(dis, True)
        if rot == "random":
            self.orientation.change_orientation()
        else:
            self.orientation.change_orientation(angle=rot / 180 * np.pi)

    def translate(self, disp=np.zeros(3), absolute=False):
        """
        To translate the molecule
        """
        disp = np.array(disp)
        if absolute:
            disp = disp.dot(self.lattice.inv_matrix)
        position = self.position + disp
        self.position = self.wp.project(position)

    def rotate(self, ax_id=0, ax_vector=None, angle=180):
        """
        To rotate the molecule
        Args:
            ax_id: the principle axis id
            ax_vector (float): 3-vector to define the axis
            angle (float): angle to rotate
        """
        p = self.orientation.r

        if ax_vector is not None:
            ax = ax_vector / np.linalg.norm(ax_vector)
        else:
            xyz = self.molecule.mol.cart_coords.dot(p.as_matrix().T)
            ax = self.molecule.get_principle_axes(xyz).T[ax_id]

        q = R.from_rotvec(ax * rad * angle)
        o = q * p
        self.orientation.r = o
        self.orientation.matrix = o.as_matrix()

    # def is_compatible_symmetry(self, tol=0.3):
    #    """
    #    Check if the molecular symmetry matches the site symmetry
    #    """
    #    mol = self.molecule.mol
    #    if len(mol) == 1 or self.wp.index==0:
    #        return True
    #    else:
    #        pga = PointGroupAnalyzer(mol, tol)
    #        for op in self.wp.symmetry_m[0]:
    #            if not pga.is_valid_op(op):
    #                return False
    #        return True

    def is_valid_orientation(self):
        pass

    def get_mol_object(self, id=0):
        """
        make the pymatgen molecule object

        Args:
            id: the index of molecules in the given site

        Returns:
            a molecule object
        """
        coord0 = self.molecule.mol.cart_coords
        coord0 = coord0 @ self.orientation.matrix.T

        # Obtain the center in absolute coords
        if not hasattr(self.wp, "generators"):
            self.wp.set_generators()

        if id <= len(self.wp.generators):
            # op = self.wp.generators[id]
            op = self.wp.get_euclidean_generator(self.lattice.matrix, id)
            center_relative = op.operate(self.position)
            center_relative -= np.floor(center_relative)
            center_absolute = np.dot(center_relative, self.lattice.matrix)

            # Rotate the molecule (Euclidean metric)
            # op_m = self.wp.generators_m[id]
            # rot = op_m.affine_matrix[0:3][:, 0:3].T
            # tau = op_m.affine_matrix[0:3][:, 3]
            op0 = self.wp.get_euclidean_generator(self.lattice.matrix, id)
            rot = op0.rotation_matrix.T
            tmp = np.dot(coord0, rot)
            # Add absolute center to molecule
            tmp += center_absolute
            return Molecule(self.symbols, tmp)
        else:
            raise ValueError("id is greater than the number of molecules")

    def show_molecule_in_box(self, id=0):
        """
        display molecule with box

        Args:
            id (int): molecular id
        """
        from pyxtal.viz import display_molecule

        mol = self.get_mol_object(id)
        cell, vertices, center = self.molecule.get_box_coordinates(
            mol.cart_coords)
        return display_molecule(mol, center, cell)

    def update(self, coords=None, lattice=None, absolute=False, update_mol=True):
        """
        After the geometry relaxation, the returned atomic coordinates
        maybe rescaled to [0, 1] bound. In this case, we need to refind
        the molecular coordinates according to the original neighbor list.
        If the list does not change, we return the new coordinates
        otherwise, terminate the calculation.

        Args:
            coords (float): input xyz coordinates for the given molecule
            absolute (bool): whether the input is absolute coordinates
            update_mol (bool): whether update the molecule (used for substitution)
        """
        from pyxtal.molecule import Orientation, compare_mol_connectivity

        try:
            from openbabel import openbabel, pybel
        except:
            import openbabel
            import pybel

        if lattice is not None:
            self.lattice = lattice

        if coords is None:
            coords, _ = self._get_coords_and_species(
                absolute=False, first=True, unitcell=True)
            absolute = False

        if not absolute:
            coords = coords.dot(self.lattice.matrix)

        # mol = Molecule(self.symbols, coords-np.mean(coords, axis=0))
        center = self.molecule.get_center(coords)
        mol = Molecule(self.symbols, coords - center)

        # Update the orientation matrix
        match, _ = compare_mol_connectivity(mol, self.molecule.mol,
                                            ignore_HH=True)
        if match:
            # position = np.mean(coords, axis=0).dot(self.lattice.inv_matrix)
            position = center.dot(self.lattice.inv_matrix)
            self.position = position - np.floor(position)
            if update_mol:
                self.orientation = Orientation(np.eye(3))
                self.molecule.mol = mol
            else:
                m1 = pybel.readstring("xyz", self.molecule.mol.to("xyz"))
                m2 = pybel.readstring("xyz", mol.to("xyz"))
                aligner = openbabel.OBAlign(True, False)
                aligner.SetRefMol(m1.OBMol)
                aligner.SetTargetMol(m2.OBMol)
                if aligner.Align():
                    print("RMSD: ", aligner.GetRMSD())
                    rot = np.zeros([3, 3])
                    for i in range(3):
                        for j in range(3):
                            rot[i, j] = aligner.GetRotMatrix().Get(i, j)
                    if abs(np.linalg.det(rot) - 1) < 1e-2:
                        self.orientation.matrix = rot
                        self.orientation.r = R.from_matrix(rot)
                    else:
                        raise ValueError("rotation matrix is wrong")
        else:
            #import pickle
            #with open("wrong.pkl", "wb") as f:
            #    pickle.dump([mol, self.molecule.mol], f)
            mol.to(filename="Wrong.xyz", fmt="xyz")
            self.molecule.mol.to(filename="Ref.xyz", fmt="xyz")
            raise ValueError("molecular connectivity changes! Exit")
        # todo check if connectivty changed

    def _create_matrix(self, center=False, ignore=False):
        """
        Used for calculating distances in lattices with periodic boundary
        conditions. When multiplied with a set of points, generates additional
        points in cells adjacent to and diagonal to the original cell

        Args:
            center:
            ignore:

        Returns:
            A numpy array of matrices which can be multiplied by a set of
            coordinates
        """
        abc = [self.lattice.a, self.lattice.b, self.lattice.c]

        # QZ: This should be based on the occupation of current molecule
        if hasattr(self, "ijk_lists"):
            ijk_lists = self.ijk_lists
        else:

            def get_ijk_range(pbc, abc_val, ignore, radius):
                if not pbc:
                    return [0]
                if not ignore and abc_val > 50.0 and radius < 10:
                    return [0]
                if abc_val < 8.5:
                    return list(range(-3, 4))
                return [-1, 0, 1]

            ijk_lists = [get_ijk_range(
                self.PBC[idx], abc[idx], ignore, self.radius) for idx in range(3)]

        matrix = [[0, 0, 0]] if center else []
        matrix += [
            [i, j, k] for i in ijk_lists[0] for j in ijk_lists[1] for k in ijk_lists[2] if [i, j, k] != [0, 0, 0]
        ]

        # In case a,b,c are all greater than 20
        if len(matrix) == 0:
            matrix = [[1, 0, 0]]
        return np.array(matrix, dtype=float)

    def get_distances(self, coord1, coord2, m2=None, center=True, ignore=False):
        """
        Compute the distance matrix between the central molecule (coord1) and
        neighboring molecules (coord2) under the periodic boundary condition.

        Args:
            coord1 (numpy array): Fractional coordinates of the central molecule.
                                Shape: (m1, 3), where m1 is the number of atoms
            coord2 (numpy array): Fractional coordinates of the neighboring molecules.
                                Shape: (N2*m2, 3), where N2 is the number of atoms
                                and m2 is the number of atoms in each neighboring molecule.
            m2 (int, optional): N_atoms in each neighboring molecule. If not provided,
                                it's assumed to be equal m1.
            center (bool, optional): If `True`, count self-image of the reference molecule
            ignore (bool, optional): If `True`, ignores some periodic boundary conditions.

        Returns:
            distance matrix: [m1*m2*pbc, m1, m2]
            coord2 under PBC: [pbc, m2, 3]
        """
        m1 = len(coord1)
        if m2 is None: m2 = m1
        N2 = int(len(coord2) / m2) # Number of molecule 2

        # peridoic images
        m = self._create_matrix(center, ignore)  # PBC matrix
        coord2 = np.vstack([coord2 + v for v in m])
        N = N2 * len(m) # Number of PBC images

        # absolute xyz
        coord1 = coord1 @ self.lattice.matrix
        coord2 = coord2 @ self.lattice.matrix

        d = cdist(coord1, coord2)
        d = d.reshape(m1, N, m2).transpose(1, 0, 2)
        coord2 = coord2.reshape([N, m2, 3])
        return d, coord2

    def get_dists_auto(self, matrix=None, ignore=False, cutoff=None):
        """
        Compute the distances between the periodic images
        N: number of atoms in the molecule
        M: number of periodic images

        Args:
            ignore (bool, optional): If `True`, ignores some periodic boundary conditions.
            cutoff (): if not None, reduce the distance matrix

        Returns:
            a distance matrix (M, N, N)
            list of molecular xyz (M, N, 3)
        """
        coord1, _ = self._get_coords_and_species(first=True, unitcell=True, matrix=matrix)
        ds, coords = self.get_distances(coord1, coord1, center=False, ignore=ignore)
        if cutoff is None:
            return ds, coords
        else:
            return self.extract_short_distances(ds, coords, cutoff, self.type, self.type)

    def extract_short_distances(self, ds, coords, cutoff, label1, label2):

        indices = np.where( ds < cutoff )
        d1s = np.zeros((len(indices[0]), 4))
        ids = np.zeros((len(indices[0]), 4), dtype=int)
        for i in range(len(indices[0])):
            m, n1, n2 = indices[0][i], indices[1][i], indices[2][i]
            d1s[i, :3] = coords[m, n2, :]
            d1s[i, 3] = ds[m, n1, n2]
            ids[i] = [label1, label2, n1, n2]
        return d1s, ids


    def get_dists_WP(self, matrix=None, ignore=False, idx=None, cutoff=None):
        """
        Compute the distances within the WP site

        Returns:
            a distance matrix (M, N, N)
            list of molecular xyz (M, N, 3)
        """
        m_length = len(self.symbols)
        coords, _ = self._get_coords_and_species(unitcell=True, matrix=matrix)
        coord1 = coords[:m_length]  # 1st molecular coords
        if idx is None:
            coord2 = coords[m_length:]
        else:
            coord2 = coords[m_length *(idx): m_length * (idx + 1)]
        ds, coords = self.get_distances(coord1, coord2, ignore=ignore)
        if cutoff is None:
            return ds, coords
        else:
            return self.extract_short_distances(ds, coords, cutoff, self.type, self.type)

    def get_dists_WP2(self, wp2, matrix=None, ignore=False, cutoff=None):
        """
        Compute the distances w.r.t to other WP site

        Returns:
            a distance matrix (M, N, N)
            list of molecular xyz (M, N, 3)
        """
        coord1, _ = self._get_coords_and_species(first=True, unitcell=True, matrix=matrix)
        coord2, _ = wp2._get_coords_and_species(unitcell=True)
        ds, coords = self.get_distances(coord1, coord2, len(wp2.numbers), ignore=ignore)
        if cutoff is None:
            return ds, coords
        else:
            return self.extract_short_distances(ds, coords, cutoff, self.type, wp2.type)

    def short_dist(self):
        """
        Check if the atoms are too close within the WP.

        Returns:
            True or False
        """
        tols_matrix = self.tols_matrix
        # Check periodic images
        d, _ = self.get_dists_auto()
        if np.min(d) < np.max(tols_matrix):
            tols = np.min(d, axis=0)
            if (tols < tols_matrix).any():
                return True

        if self.wp.multiplicity > 1:
            d, _ = self.get_dists_WP()
            if np.min(d) < np.max(tols_matrix):
                tols = np.min(d, axis=0)  # N*N matrix
                if (tols < tols_matrix).any():
                    return True

        return False

    def short_dist_with_wp2(self, wp2, tm=Tol_matrix(prototype="molecular")):
        """
        Check whether or not the molecules of two wp sites overlap. Uses
        ellipsoid overlapping approximation to check.

        Args:
            wp2: the 2nd wp sites
            tm: a Tol_matrix object (or prototype string) for distance checking

        Returns:
            True or False
        """

        # Get coordinates for both mol_sites
        c1, _ = self.get_coords_and_species()
        c2, _ = wp2.get_coords_and_species()
        m_length1 = len(self.numbers)
        m_length2 = len(wp2.numbers)

        # choose wp with bigger molecule as the center
        if len(c2) <= len(c1):
            coord1 = c1[:m_length1]
            coord2 = c2  # rest molecular coords
            tols_matrix = self.molecule.get_tols_matrix(wp2.molecule, tm)
            m2 = m_length2
        else:
            coord1 = c2[:m_length2]
            coord2 = c1
            tols_matrix = wp2.molecule.get_tols_matrix(self.molecule, tm)
            m2 = m_length1

        # compute the distance matrix
        d, _ = self.get_distances(
            coord1 - np.floor(coord1), coord2 - np.floor(coord2), m2)
        # print("short dist", len(c1), len(c2), d.min())

        if np.min(d) < np.max(tols_matrix):
            tols = np.min(d, axis=0)
            if (tols < tols_matrix).any():
                return False
        return True

    def get_neighbors_auto(self, factor=1.1, max_d=4.0, ignore_E=True, detail=False, etol=-5e-2):
        """
        Find neighboring molecules within a given distance threshold.

        The function identifies neighboring molecules within PBC and computes the shortest
        distances and (optionally) interaction energies. it returns detailed information
        about neighboring molecule pairs, distances, and energies.

        Args:
            factor (float, optional): Scaling factor for distance (default is 1.1).
            max_d (float, optional): Maximum distance for neighbors (default is 4.0 Å).
            ignore_E (bool, optional): Skips energy calculations (default is `True`).
            detail (bool, optional): Returns detailed eng, molecular pairs and
                                     distances instead of only shortest distances.
            etol (float, optional): Energy tolerance for filtering pairs in detailed
                                    mode (default is -5e-2).

        Returns:
            If `detail is True`:
                engs (list): List of interaction energies for valid molecular pairs.
                pairs (list): List of tuples with neighbor molecules and positions.
                dists (list): List of distances between neighboring molecular pairs.
            If `detail is False`:
                min_ds (list): List of shortest distances between central and neighbors.
                neighs (list): List of neighboring molecular coordinates with PBC.
                Ps (list): List of Wyckoff position multiplicities or translations.
                engs (list): List of energies, if energy calculation is not skipped.
        """

        # Compute the mol_center in Cartesian coordinate
        position = self.position - np.floor(self.position)
        mol_center = np.dot(position, self.lattice.matrix)

        # Atomic numbers for atoms in the central molecule
        numbers = self.molecule.mol.atomic_numbers

        # Get fractional coordinates for the central molecule
        coord1 = self._get_coords_and_species(first=True, unitcell=True)[0]

        # Initialize tolerance matrix for intermolecular distances
        tm = Tol_matrix(prototype="vdW", factor=factor)
        tols_matrix = self.molecule.get_tols_matrix(tm=tm)

        # Initialize coefficient matrix for energy calculations if needed
        coef_matrix = None
        if not ignore_E:
            coef_matrix = self.molecule.get_coefs_matrix()
            if coef_matrix is not None:
                A = coef_matrix[:, :, 0]
                B = coef_matrix[:, :, 1]
                C = coef_matrix[:, :, 2]

        # Initialize lists for results
        min_ds = []
        neighs = []
        Ps = []
        engs = []
        pairs = []
        dists = []

        # Find neighbors under PBC
        d, coord2 = self.get_dists_auto(ignore=True)

        # Loop through each neighboring molecule
        for i in range(d.shape[0]):
            if np.min(d[i]) < max_d and (d[i] < tols_matrix).any():
                if coef_matrix is not None:
                    if detail:
                        eng = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                        ids = np.where(eng < etol)
                        for id in range(len(ids[0])):
                            n1, n2 = numbers[ids[0][id]], numbers[ids[1][id]]
                            if 1 not in [n1, n2]:
                                pos = coord2[i][ids[1][id]] - mol_center
                                pairs.append((n2, pos))
                                engs.append(eng[ids[0][id], ids[1][id]])
                                dists.append(d[i][ids[0][id], ids[1][id]])
                    else:
                        eng0 = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                        engs.append(eng0.sum())
                else:
                    engs.append(None)
                    if detail:
                        ids = np.where(d[i] < max_d)
                        for id in range(len(ids[0])):
                            n1, n2 = numbers[ids[0][id]], numbers[ids[1][id]]
                            if 1 not in [n1, n2]:
                                pos = coord2[i][ids[1][id]] - mol_center
                                pairs.append((n2, pos))
                                dists.append(np.linalg.norm(pos))

                tmp = d[i] / tols_matrix
                _d = tmp[tmp < 1.0]
                id = np.argmin(tmp.flatten())
                d[i].flatten()[id]
                min_ds.append(min(_d) * factor)
                neighs.append(coord2[i])
                Ps.append(0)

        # Handle Wyckoff position multiplicities (if applicable)
        if self.wp.multiplicity > 1:
            for idx in range(1, self.wp.multiplicity):
                P = 0 if self.wp.is_pure_translation(idx) else 1
                d, coord2 = self.get_dists_WP(ignore=True, idx=idx)
                for i in range(d.shape[0]):
                    if np.min(d[i]) < max_d and (d[i] < tols_matrix).any():
                        if coef_matrix is not None:
                            if detail:
                                eng = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                                ids = np.where(eng < etol)
                                for id in range(len(ids[0])):
                                    n1, n2 = numbers[ids[0][id]
                                                     ], numbers[ids[1][id]]
                                    if 1 not in [n1, n2]:
                                        pos = coord2[i][ids[1]
                                                        [id]] - mol_center
                                        pairs.append((n2, pos))
                                        engs.append(
                                            eng[ids[0][id], ids[1][id]])
                                        dists.append(
                                            d[i][ids[0][id], ids[1][id]])

                            else:
                                eng0 = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                                engs.append(eng0.sum())
                        else:
                            engs.append(None)
                            if detail:
                                # print('OMMMM', d[i].min())
                                ids = np.where(d[i] < max_d)
                                # for id in zip(*ids):
                                for id in range(len(ids[0])):  # zip(*ids):
                                    n1, n2 = numbers[ids[0][id]
                                                     ], numbers[ids[1][id]]
                                    if 1 not in [n1, n2]:  # != [1, 1]:
                                        pos = coord2[i][ids[1]
                                                        [id]] - mol_center
                                        pairs.append((n2, pos))
                                        dists.append(np.linalg.norm(pos))

                        tmp = d[i] / tols_matrix
                        _d = tmp[tmp < 1]
                        id = np.argmin(tmp.flatten())
                        d[i].flatten()[id]
                        min_ds.append(min(_d) * factor)
                        neighs.append(coord2[i])
                        Ps.append(P)

        # Return results based on the detail flag
        if detail:
            return engs, pairs, dists
        else:
            return min_ds, neighs, Ps, engs

    def get_neighbors_wp2(self, wp2, factor=1.1, max_d=4.0, ignore_E=True, detail=False, etol=-5e-2):
        """
        Find the neigboring molecules from a 2nd wp site

        Returns
            min_ds: list of shortest distances
            neighs: list of neighboring molecular xyzs
        """

        tm = Tol_matrix(prototype="vdW", factor=factor)
        m_length1 = len(self.numbers)
        m_length2 = len(wp2.numbers)

        # Get coordinates for both mol_sites
        c1, _ = self.get_coords_and_species()
        c2, _ = wp2.get_coords_and_species()

        coord1 = c1[:m_length1]
        coord2 = c2  # rest molecular coords
        tols_matrix = self.molecule.get_tols_matrix(wp2.molecule, tm)
        coef_matrix = None
        if not ignore_E:
            coef_matrix = self.molecule.get_coefs_matrix(wp2.molecule)
            if coef_matrix is not None:
                A = coef_matrix[:, :, 0]
                B = coef_matrix[:, :, 1]
                C = coef_matrix[:, :, 2]

        # compute the distance matrix
        d, coord2 = self.get_distances(coord1, coord2, m_length2, ignore=True)
        min_ds = []
        neighs = []
        engs = []
        dists = []
        pairs = []

        for i in range(d.shape[0]):
            if np.min(d[i]) < max_d and (d[i] < tols_matrix).any():
                if coef_matrix is not None:
                    if detail:
                        eng = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                        ids = np.where(eng < etol)
                        for id in zip(*ids):
                            tmp1, tmp2 = coord1[id[0]], coord2[i][id[1]]
                            pairs.append((tmp1 + tmp2) / 2)
                            engs.append(eng[id])
                            dists.append(d[i][id])
                    else:
                        eng0 = A * np.exp(-B * d[i]) - C / (d[i] ** 6)
                        engs.append(eng0.sum())
                else:
                    engs.append(None)
                tmp = d[i] / tols_matrix
                _d = tmp[tmp < 1]
                id = np.argmin(tmp.flatten())
                d[i].flatten()[id]
                min_ds.append(min(_d) * factor)
                neighs.append(coord2[i])

        if detail:
            return engs, pairs, dists
        else:
            return min_ds, neighs, engs

    def get_ijk_lists(self, value=None):
        """
        Get the occupatation in the unit cell for the generating molecule
        This can be used to estimate the supercell size for finding neighbors

        Returns
            PBC
        """

        # Get coordinates for both mol_sites
        if value is None:
            ijk_lists = []
            c1, _ = self.get_coords_and_species(absolute=False)
            for id in range(3):
                if self.PBC[id]:
                    m1 = np.min(c1[:, id])
                    m2 = np.max(c1[:, id])
                    max_id = int(np.ceil(2 * m2 - m1))
                    min_id = int(np.floor(-m2))
                    ijk_lists.append(list(range(min_id - 1, max_id + 1)))
                else:
                    ijk_lists.append([0])
            self.ijk_lists = ijk_lists
        else:
            self.ijk_lists = value

    def to_atom_site(self, specie=1):
        """
        transform it to the mol_sites, i.e., to build a molecule on
        the current WP
        """
        dicts = {}
        dicts["specie"] = specie
        dicts["position"] = self.position
        dicts["hn"] = self.wp.hall_number
        dicts["index"] = self.wp.index
        return atom_site.load_dict(dicts)


if __name__ == "__main__":

    from pyxtal.symmetry import Wyckoff_position as WP

    wp = WP.from_group_and_letter(225, 'a')
    coordinate = [0.25, 0.25, 0.25]
    specie = 6
    atom_site_instance = atom_site(wp=wp,
                                   coordinate=coordinate,
                                   specie=specie)

    # Print the created instance
    print(atom_site_instance)



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import pymatgen.analysis.structure_matcher as sm

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.supergroup import supergroup, supergroups
from pyxtal.symmetry import Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestSupergroup(unittest.TestCase):
    def test_supergroup(self):
        """
        call supergroup from pyxtal
        """
        data = [
            ("NbO2", 141, 5),
        ]
        for d in data:
            (cif, g, N) = d
            s = pyxtal()
            s.from_seed(cif_path + cif + ".cif")
            strucs, sols = s.supergroup(g, 0.5)
            assert len(strucs) > 0

    def test_make_pyxtal(self):
        data = {
            # "NbO2": 141,
            "GeF2": 62,
            "lt_quartz": 180,
            "NiS-Cm": 160,  # 9b->2a+4b
        }
        for cif in data:
            s = pyxtal()
            s.from_seed(cif_path + cif + ".cif")
            my = supergroup(s, G=data[cif])
            sols = my.search_supergroup(max_solutions=1)
            for sol in sols:
                struc_high = my.make_pyxtal_in_supergroup(sol)
                strucs = my.make_pyxtals_in_subgroup(sol, 3)
                pmg1 = struc_high.to_pymatgen()
                pmg2 = strucs[-1].to_pymatgen()
                rms = sm.StructureMatcher().get_rms_dist(pmg1, pmg2)[0]
                print(cif, rms)
                assert rms < 0.001

    def test_by_group(self):
        data = {
            "BTO-Amm2": 221,
            "BTO": 221,
            #"lt_cristobalite": 227,
            "NaSb3F10": 194,
        }
        for cif in data:
            s = pyxtal()
            s.from_seed(cif_path + cif + ".cif")
            sup = supergroups(s, G=data[cif], show=False)
            assert sup.strucs is not None

    def test_by_path(self):
        data = {
            "MPWO": [59, 71, 139, 225],
        }
        for cif in data:
            s = pyxtal()
            s.from_seed(cif_path + cif + ".cif")
            sup = supergroups(s, path=data[cif], show=False)
            assert sup.strucs is not None

    def test_long(self):
        paras = (
            ["I4_132", 98],  # 1-3
            ["P3_112", 5],  # 1-3
            ["P6_422", 21],  # 1-3
        )
        for para in paras:
            name, H = para
            name = cif_path + name + ".cif"
            s = pyxtal()
            s.from_seed(name)
            pmg_s1 = s.to_pymatgen()
            G = s.group.number
            struc_h = s.subgroup_once(eps=0.05, H=H, mut_lat=False)
            sup = supergroups(struc_h, G=G, d_tol=0.2, show=False, max_per_G=500)

            if sup.strucs is None:
                print("Problem in ", name)
            assert sup.strucs is not None

            match = False
            for struc in sup.strucs:
                pmg_g = struc.to_pymatgen()
                if sm.StructureMatcher().fit(pmg_g, pmg_s1):
                    match = True
                    break
            if not match:
                print("Problem in ", name)
            assert match

    def test_long2(self):
        paras = (
            ["P4_332", 155],  # 1-4 splitting
            #["Fd3", 70],  # 1-3
            ["Pm3", 47],  # 1-3
            ["Fd3m", 166],
            ["R-3c", 15],
            #["R32", 5],
            ["R-3", 147],  # 1-3, k
            ["P4_332", 96],  # 1-3
        )
        for para in paras:
            name, H = para
            name = cif_path + name + ".cif"
            s = pyxtal()
            s.from_seed(name)
            pmg_s1 = s.to_pymatgen()
            G = s.group.number
            struc_h = s.subgroup_once(eps=0, H=H, mut_lat=False)
            my = supergroup(struc_h, G=G)
            sols = my.search_supergroup(max_solutions=1)

            if len(sols) == 0:
                print("problem", name)
            assert len(sols) > 0

            struc_high = my.make_pyxtal_in_supergroup(sols[0])
            struc_sub = my.make_pyxtals_in_subgroup(sols[0], 3)[-1]

            pmg_high = struc_high.to_pymatgen()
            pmg_sub = struc_sub.to_pymatgen()

            dist1 = sm.StructureMatcher().get_rms_dist(pmg_high, pmg_sub)[0]
            dist2 = sm.StructureMatcher().get_rms_dist(pmg_s1, pmg_high)[0]
            assert dist1 < 0.001
            assert dist2 < 0.001

    def test_multi(self):
        data = {
            "BTO": [123, 221],
            "lt_cristobalite": [98, 210, 227],
            #"BTO-Amm2": [65, 123, 221],
            # "NaSb3F10": [186, 194],
            # "NaSb3F10": [176, 194],
            # "MPWO": [59, 71, 139, 225],
        }
        for cif in data:
            s = pyxtal()
            s.from_seed(cif_path + cif + ".cif")
            sup = supergroups(s, path=data[cif], show=False, max_per_G=2500)
            strucs = sup.get_transformation()
            pmg_0, pmg_1 = s.to_pymatgen(), sup.strucs[-1].to_pymatgen()
            pmg_2, pmg_3 = strucs[0].to_pymatgen(), strucs[1].to_pymatgen()
            dist1 = sm.StructureMatcher().get_rms_dist(pmg_0, pmg_2)[0]
            dist2 = sm.StructureMatcher().get_rms_dist(pmg_1, pmg_3)[0]
            print(cif, dist1, dist2)
            if dist2 > 1e-3:
                print(pmg_1)
                print(pmg_3)
            assert dist1 < 0.001
            assert dist2 < 0.001

    def test_similarity(self):
        paras = [
            ("0-G62", "2-G71"),
            ("0-G62", "3-G139"),
            # ('0-G62', '4-G225'),
        ]
        for para in paras:
            (cif1, cif2) = para
            s1 = pyxtal()
            s2 = pyxtal()
            s1.from_seed(cif_path + cif1 + ".cif")
            s2.from_seed(cif_path + cif2 + ".cif")
            pmg_s2 = s2.to_pymatgen()

            strucs, _, _, _, _ = s2.get_transition(s1)

            if strucs is None:
                print("Problem between ", cif1, cif2)
            else:
                struc_G_in_H = strucs[-1]
                s3 = pyxtal()
                s3.from_seed(struc_G_in_H.to_pymatgen(), tol=1e-3)
                pmg_s3 = s3.to_pymatgen()
                dist = sm.StructureMatcher().get_rms_dist(pmg_s2, pmg_s3)[0]
                assert dist < 0.001
                assert s3.group.number == s2.group.number

    def test_get_disp_sets(self):
        s1 = pyxtal()
        s1.from_seed(cif_path + "dist_6_0.cif")
        s2 = pyxtal()
        s2.from_seed(cif_path + "dist_6_1.cif")
        _, _, _, d = s1.get_disps_sets(s2, 1.0)
        assert d < 0.15

        #s1 = pyxtal()
        #s1.from_seed(cif_path + "sim-0.vasp")
        #s2 = pyxtal()
        #s2.from_seed(cif_path + "sim-1.vasp")
        #_, _, _, d = s1.get_disps_sets(s2, 1.0, 0.3)
        #assert d < 0.02

if __name__ == "__main__":
    unittest.main()



"""
Module for generating crystals based on building blocks
"""
# Standard Libraries
import numpy as np

# External Libraries
from pymatgen.core import Molecule

# PyXtal imports
from pyxtal.molecular_crystal import molecular_crystal as mol_xtal
from pyxtal.molecule import Orientation, compare_mol_connectivity, pyxtal_molecule
from pyxtal.wyckoff_site import mol_site
from pyxtal.io import search_molecules_in_crystal

def block_crystal(
    dim,
    group,
    molecules,
    numMols,
    factor,
    thickness,
    area,
    block,
    num_block,
    lattice,
    torsions,
    sites,
    conventional,
    tm,
    seed,
    random_state,
    use_hall,
):
    # If block is None, directly generate mol. xtal.
    # Otherwise, generate crystal from building block
    if block is None:
        return mol_xtal(
            dim,
            group,
            molecules,
            numMols,
            factor,
            thickness=thickness,
            area=area,
            lattice=lattice,
            torsions=torsions,
            sites=sites,
            conventional=conventional,
            tm=tm,
            seed=seed,
            random_state=random_state,
            use_hall=use_hall,
        )

    else:
        p_mol = pyxtal_molecule(block)
        block_mols = search_molecules_in_crystal(p_mol.mol, tol=0.2, once=False)
        xtal_mols = [pyxtal_molecule(m, fix=True) for m in molecules]

        orders = []
        for m1 in xtal_mols:
            for m2 in block_mols:
                if len(m1.mol) == len(m2):
                    # match, mapping = compare_mol_connectivity(m2, m1.mol)
                    match, mapping = compare_mol_connectivity(m1.mol, m2)
                    # print(match, len(m1.mol), len(m2))
                    if match:
                        orders.append([mapping[at] for at in range(len(m2))])
                        break
                        # print(mapping)
                        # print("smile"); print(m1.mol.to('xyz'))
                        # print("reference"); print(m2.to('xyz'))
                        # print(orders[-1])
                        # import sys; sys.exit()

        if len(orders) != len(molecules):
            raise ValueError("Block is inconsistent with the molecules")

        # rearrange the order of block molecules
        numbers = []
        coords = np.zeros([len(p_mol.mol), 3])
        count = 0
        for order, m in zip(orders, block_mols):
            numbers.extend([m.atomic_numbers[o] for o in order])
            coords[count : count + len(m)] += m.cart_coords[order]
            count += len(m)
        mol = Molecule(numbers, coords)
        if num_block is not None:
            num_block = [num_block]
        # print(mol.to('xyz')); import sys; sys.exit()

        for i in range(10):
            struc = mol_xtal(
                dim,
                group,
                [mol],
                num_block,
                factor,
                thickness=thickness,
                area=area,
                lattice=lattice,
                torsions=torsions,
                sites=sites,
                conventional=conventional,
                tm=tm,
                random_state=random_state,
                use_hall=use_hall,
            )

            if struc.valid:
                break

        struc.molecules = xtal_mols
        struc.numMols = struc.numMols * len(xtal_mols)

        # XYZ from the block_sites
        b_site = struc.mol_sites[0]
        xyz, _ = b_site._get_coords_and_species(absolute=True, first=True)

        # Create mol sites
        ori = Orientation(np.eye(3))
        mol_sites = []

        count = 0
        for i in range(len(molecules)):
            mol = xtal_mols[i]
            m_xyz = xyz[count : count + len(mol.mol)]
            center = mol.get_center(m_xyz)
            mol.reset_positions(m_xyz - center)
            # print(mol.mol.to('xyz'))
            position = np.dot(center, np.linalg.inv(struc.lattice.matrix))
            position -= np.floor(position)
            m_site = mol_site(mol, position, ori, b_site.wp, struc.lattice)
            m_site.type = i
            mol_sites.append(m_site)
            count += len(mol.mol)

        struc.mol_sites = mol_sites

        return struc


if __name__ == "__main__":
    import pymatgen.analysis.structure_matcher as sm

    from pyxtal import pyxtal
    from pyxtal.representation import representation

    smiles = [
        "C1=C(C=C(C=C1[N+](=O)[O-])[N+](=O)[O-])C(=O)O.smi",
        "CC1=CC2=C(C=C1)N3CC4=C(C=CC(=C4)C)N(C2)C3.smi",
    ]

    for block in [None, "xxv"]:
        print("block", block)
        for i in range(30):
            s = pyxtal(molecular=True)
            s.from_random(3, 14, smiles, block=block)

            rep2 = representation.from_pyxtal(s)
            strs = rep2.to_string()
            print(i, strs)

            rep3 = representation.from_string(strs, smiles)
            s1 = rep3.to_pyxtal()

            pmg1 = s.to_pymatgen()
            pmg2 = s1.to_pymatgen()
            pmg1.remove_species("H")
            pmg2.remove_species("H")
            if not sm.StructureMatcher().fit(pmg1, pmg2):
                print("Mismatch")
                print("S1", s.check_short_distances())
                print("S2", s1.check_short_distances())
                s.to_file("1.cif")
                s1.to_file("2.cif")
                import sys; sys.exit()


# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import pymatgen.analysis.structure_matcher as sm
from pymatgen.core import Structure

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.symmetry import Hall, Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestDof(unittest.TestCase):
    def test_atomic(self):
        s = pyxtal()
        s.from_random(3, 225, ["C"], [8])
        ans = s.get_dof()
        assert s.lattice.dof == 1
        assert ans == 1


class TestAtomic3D(unittest.TestCase):
    def test_single_specie(self):
        struc = pyxtal()
        struc.from_random(3, 225, ["C"], [4], 1.2, conventional=False)
        struc.to_file("tmp-3d.cif")
        os.remove("tmp-3d.cif")
        assert struc.valid

    def test_mutiple_species(self):
        struc = pyxtal()
        struc.from_random(3, 99, ["Ba", "Ti", "O"], [1, 1, 3], 1.2)
        assert struc.valid

    def test_preassigned_sites(self):
        sites = [["1b"], ["1b"], ["2c", "1b"]]
        struc = pyxtal()
        struc.from_random(3, 99, ["Ba", "Ti", "O"], [1, 1, 3], 1.0, sites=sites)
        assert struc.valid

        struc = pyxtal()
        struc.from_random(3, 225, ["C"], [12], 1.0, sites=[["4a", "8c"]])
        assert struc.valid

    def test_read(self):
        # test reading xtal from cif
        for name in ["FAU", "NaSb3F10", "PVO", "lt_quartz"]:
            cif_file = cif_path + name + ".cif"
            pmg1 = Structure.from_file(cif_file, primitive=True)
            struc = pyxtal()
            struc.from_seed(seed=cif_file)
            pmg_struc = struc.to_pymatgen()
            assert sm.StructureMatcher().fit(pmg_struc, pmg1)

    def test_read_spglib(self):
        # test reading xtal from cif
        for name in ["FAU"]:
            cif_file = cif_path + name + ".cif"
            pmg1 = Structure.from_file(cif_file, primitive=True)
            struc = pyxtal()
            struc.from_seed(seed=cif_file, style="spglib")
            pmg_struc = struc.to_pymatgen()
            assert sm.StructureMatcher().fit(pmg_struc, pmg1)
        # more space groups
        for name in ["I41amd", "P4nmm", "Pmmn", "Pn3m", "Fd3", "Pn3"]:
            cif_file = cif_path + name + ".vasp"
            pmg1 = Structure.from_file(cif_file, primitive=True)
            struc = pyxtal()
            struc.from_seed(seed=cif_file, style="spglib")
            pmg_struc = struc.to_pymatgen()
            assert sm.StructureMatcher().fit(pmg_struc, pmg1)

    def test_read_by_HN(self):
        for name in ["aspirin"]:
            cif_file = cif_path + name + ".cif"
            pmg1 = Structure.from_file(cif_file, primitive=True)
            struc = pyxtal()
            for hn in Hall(14).hall_numbers:
                struc._from_pymatgen(pmg1, hn=hn)
                pmg_struc = struc.to_pymatgen()
                assert sm.StructureMatcher().fit(pmg_struc, pmg1)

    def test_from_tabular(self):
        xtal = pyxtal()
        rep = [116,10.5754,10.7203,4.47208,1.5705,2.6561,2.0943,1,0.4447,0.3762,0.7526]
        xtal.from_tabular_representation(rep, normalize=False)
        assert xtal.valid == False
        rep = [116,10.5754,10.7203,4.47208,1.5705,1.5705,1.5705,0,0.4447,0.3762,0.7526]
        xtal.from_tabular_representation(rep, normalize=False)
        assert xtal.valid == True

        rep0 = xtal.get_tabular_representation(discrete_cell=True,
                                               discrete=True,
                                               N_grids=100)
        assert(int(rep0[6]) == 50)

        xtal.from_tabular_representation(rep0,
                                         discrete_cell=True,
                                         discrete=True,
                                         N_grids=100)
        assert(int(xtal.lattice.get_para(degree=True)[-1]) == 90)

        reps = xtal.get_tabular_representations(N_wp=1,
                                                discrete_cell=True,
                                                discrete=True,
                                                N_grids=100)
        assert(len(reps)==8)

class TestAtomic2D(unittest.TestCase):
    def test_single_specie(self):
        struc = pyxtal()
        struc.from_random(2, 20, ["C"], [4], 1.0, thickness=2.0)
        struc.to_file("tmp-2d.cif")
        os.remove("tmp-2d.cif")
        assert struc.valid

    def test_mutiple_species(self):
        struc = pyxtal()
        struc.from_random(2, 4, ["Mo", "S"], [2, 4], 1.0)
        assert struc.valid


class TestAtomic1D(unittest.TestCase):
    def test_single_specie(self):
        struc = pyxtal()
        struc.from_random(1, 20, ["C"], [4], 1.0)
        struc.to_file("tmp-1d.cif")
        os.remove("tmp-1d.cif")
        assert struc.valid

    def test_mutiple_species(self):
        struc = pyxtal()
        struc.from_random(1, 4, ["Mo", "S"], [2, 4], 1.0)
        assert struc.valid


class TestCluster(unittest.TestCase):
    def test_multi_sites(self):
        struc = pyxtal()
        struc.from_random(0, 1, ["C"], [60], 1.0)
        assert struc.valid

        struc = pyxtal()
        struc.from_random(0, 3, ["C"], [60], 1.0)
        assert struc.valid

    def test_single_specie(self):
        struc = pyxtal()
        struc.from_random(0, "Ih", ["C"], [60], 1.0)
        assert struc.valid

    def test_mutiple_species(self):
        struc = pyxtal()
        struc.from_random(0, 4, ["Mo", "S"], [2, 4], 1.0)
        assert struc.valid

if __name__ == "__main__":
    unittest.main()
